import re
import sys
from pathlib import Path
from typing import Tuple

from pyreball.utils.logger import get_logger

logger = get_logger()


def get_html(template_path: Path, title: str, css_definitions: str) -> Tuple[str, str]:
    with open(template_path, "r") as f:
        html_text = f.read()
        html_start, html_end = html_text.split("<!--PYREBALL_REPORT_CONTENTS-->")
        html_start = re.sub("<!--PYREBALL_PAGE_TITLE-->", title, html_start)
        html_start = re.sub(
            r"<!--PYREBALL_CSS_DEFINITIONS-->", css_definitions, html_start
        )
        return html_start, html_end


def get_css(filename: str, directory: Path, page_width: int = 60) -> str:
    try:
        with open(directory / filename) as f:
            css_string = f.read()
    except OSError:
        logger.error(
            f"There was a problem reading file {filename} in {directory}. "
            f"Try re-installing pyreball."
        )
        sys.exit(1)
    css_string = re.sub(r"{{page_width}}", str(int(page_width)), css_string)
    return css_string



import configparser
import logging
import os
from pathlib import Path
from unittest import mock

import pytest

from pyreball.utils.param import (
    _map_env_value,
    _matches_paging_sizes_string,
    carefully_remove_directory_if_exists,
    check_and_fix_parameters,
    check_choice_string_parameter,
    check_integer_within_range,
    check_paging_sizes_string_parameter,
    ChoiceParameter,
    get_external_links_from_config,
    get_file_config,
    get_parameter_value,
    IntegerParameter,
    make_sure_dir_exists,
    merge_parameter_dictionaries,
    merge_values,
    read_file_config,
    StringParameter,
    Substitutor,
)

MODULE_PATH = "pyreball.utils.param"


@pytest.fixture
def simple_parameter_specifications():
    return [
        ChoiceParameter(
            "--align", choices=["left", "center", "right"], default="center", help=""
        ),
        ChoiceParameter("--do-stuff", choices=["yes", "no"], default="yes", help=""),
        ChoiceParameter("--highlight", choices=["yes", "no"], default="no", help=""),
        ChoiceParameter("--organize", choices=["yes", "no"], default="no", help=""),
        IntegerParameter("--page-width", boundaries=(10, 20), default=15, help=""),
        StringParameter(
            "--paging-sizes",
            default="10,25,100,All",
            help="",
            validation_function=check_paging_sizes_string_parameter,
        ),
    ]


@pytest.mark.parametrize(
    "test_input,expected_result",
    [
        ("", False),
        ("23", True),
        ("23,452", True),
        ("23,452,100", True),
        ("23,452,100,All", True),
        ("23,all,100", True),
        ("ALL", True),
        ("unknown", False),
        ("23,452 100", False),
        ("23_452,100", False),
        ("Allright", False),
        ("23,", False),
        (",23", False),
        (",", False),
    ],
)
def test__matches_paging_sizes_string(test_input, expected_result):
    assert _matches_paging_sizes_string(test_input) == expected_result


@pytest.mark.parametrize(
    "value,none_allowed,err_msg,expected_result",
    [
        ("10,ALL", True, None, "10,ALL"),
        ("10,ALL", False, None, "10,ALL"),
        (None, True, None, None),
        (None, False, None, "30,40"),
        ("hello", True, "Parameter param1 is set to an unsupported", "hello"),
        ("hello", False, "Parameter param1 is set to an unsupported", "hello"),
    ],
)
def test_check_paging_sizes_string_parameter(
    value, none_allowed, err_msg, expected_result
):
    error_messages = []
    result_value = check_paging_sizes_string_parameter(
        "param1",
        value,
        "30,40",
        none_allowed,
        [],
        error_messages,
    )
    assert result_value == expected_result
    contains_error_msg = any(map(lambda msg: err_msg in msg, error_messages))
    assert contains_error_msg if err_msg else not contains_error_msg


@pytest.mark.parametrize(
    "value,none_allowed,warn_msg,err_msg,expected_result",
    [
        ("one", True, None, None, "one"),
        ("one", False, None, None, "one"),
        (None, True, None, None, None),
        (None, False, "Parameter param1 was not set", None, "three"),
        ("four", True, None, "Parameter param1 is set to an unsupported", "four"),
        ("four", False, None, "Parameter param1 is set to an unsupported", "four"),
    ],
)
def test_check_choice_string_parameter(
    value, none_allowed, warn_msg, err_msg, expected_result
):
    warning_messages = []
    error_messages = []
    result_value = check_choice_string_parameter(
        "param1",
        value,
        ["one", "two"],
        "three",
        none_allowed,
        warning_messages,
        error_messages,
    )
    assert result_value == expected_result
    contains_warning_msg = any(map(lambda msg: warn_msg in msg, warning_messages))
    contains_error_msg = any(map(lambda msg: err_msg in msg, error_messages))
    assert contains_warning_msg if warn_msg else not contains_warning_msg
    assert contains_error_msg if err_msg else not contains_error_msg


@pytest.mark.parametrize(
    "value,low,high,none_allowed,warn_msg,err_msg,expected_result",
    [
        (25, 20, 30, True, None, None, 25),
        (25, 20, 30, False, None, None, 25),
        (None, 20, 30, True, None, None, None),
        (None, 20, 30, False, None, "Could not parse param1", None),
        ("four", 20, 30, True, None, "Could not parse param1", "four"),
        ("four", 20, 30, False, None, "Could not parse param1", "four"),
        (
            10,
            20,
            30,
            True,
            "Parameter param1 is less than 20, setting it to 20.",
            None,
            20,
        ),
        (
            50,
            20,
            30,
            False,
            "Parameter param1 is more than 30, setting it to 30.",
            None,
            30,
        ),
        (25, None, None, False, None, None, 25),
        (
            25,
            None,
            20,
            False,
            "Parameter param1 is more than 20, setting it to 20.",
            None,
            20,
        ),
        (
            25,
            30,
            None,
            False,
            "Parameter param1 is less than 30, setting it to 30.",
            None,
            30,
        ),
    ],
)
def test_check_integer_within_range(
    value, low, high, none_allowed, warn_msg, err_msg, expected_result
):
    warning_messages = []
    error_messages = []
    result_value = check_integer_within_range(
        "param1", value, low, high, none_allowed, warning_messages, error_messages
    )
    assert result_value == expected_result
    contains_warning_msg = any(map(lambda msg: warn_msg in msg, warning_messages))
    contains_error_msg = any(map(lambda msg: err_msg in msg, error_messages))
    assert contains_warning_msg if warn_msg else not contains_warning_msg
    assert contains_error_msg if err_msg else not contains_error_msg


@pytest.mark.parametrize(
    "parameters,none_allowed,warn_msgs,err_msgs,expected_result",
    [
        # everything is fine
        (
            {
                "align": "center",
                "do_stuff": "no",
                "highlight": "yes",
                "organize": "yes",
                "page_width": 10,
                "paging_sizes": "All,100",
            },
            True,
            [],
            [],
            {
                "align": "center",
                "do_stuff": "no",
                "highlight": "yes",
                "organize": "yes",
                "page_width": 10,
                "paging_sizes": "All,100",
            },
        ),
        # None values are used for the missing ones
        (
            {
                "do_stuff": "no",
                "highlight": "yes",
                "page_width": 30,
            },
            True,
            ["Parameter page_width is more than"],
            [],
            {
                "align": None,
                "do_stuff": "no",
                "highlight": "yes",
                "organize": None,
                "page_width": 20,
                "paging_sizes": None,
            },
        ),
        # Default values are used for the missing ones
        (
            {
                "do_stuff": "no",
                "highlight": "yes",
                "page_width": 4,
            },
            False,
            [
                "Parameter align was not set",
                "Parameter organize was not set",
                "Parameter page_width is less than",
                "Parameter paging_sizes was not set",
            ],
            [],
            {
                "align": "center",
                "do_stuff": "no",
                "highlight": "yes",
                "organize": "no",
                "page_width": 10,
                "paging_sizes": "10,25,100,All",
            },
        ),
        # Unsupported values
        (
            {
                "do_stuff": "maybe",
                "highlight": "yes",
                "organize": "no",
                "page_width": "not_a_number",
                "paging_sizes": "All,100",
            },
            False,
            ["Parameter align was not set"],
            [
                "Parameter do_stuff is set to an unsupported",
                "Could not parse page_width parameter as an integer",
            ],
            "result_does_not_matter",
        ),
    ],
)
def test_check_and_fix_parameters(
    parameters,
    none_allowed,
    warn_msgs,
    err_msgs,
    expected_result,
    caplog,
    simple_parameter_specifications,
):
    caplog.set_level(logging.WARNING)
    if len(err_msgs) > 0:
        with pytest.raises(SystemExit):
            check_and_fix_parameters(
                parameters=parameters,
                parameter_specifications=simple_parameter_specifications,
                none_allowed=none_allowed,
            )
        for msg in warn_msgs:
            assert msg in caplog.text
        for msg in err_msgs:
            assert msg in caplog.text
    else:
        result_parameters = check_and_fix_parameters(
            parameters=parameters,
            parameter_specifications=simple_parameter_specifications,
            none_allowed=none_allowed,
        )
        for msg in warn_msgs:
            assert msg in caplog.text
        assert result_parameters == expected_result


def test_read_file_config__existing_file(tmpdir):
    filename = "conf"
    directory = Path(tmpdir)
    with open(directory / filename, "w") as f:
        f.writelines(["[Parameters]\n", "a = b\n"])
    config = read_file_config(filename, directory)
    assert config["Parameters"]["a"] == "b"


def test_read_file_config__non_existing_file(tmpdir, caplog):
    caplog.set_level(logging.ERROR)
    filename = "conf"
    directory = Path(tmpdir)
    with pytest.raises(SystemExit):
        read_file_config(filename, directory)
    assert "Could not find file" in caplog.text


def test_get_file_config__correct_specification(simple_parameter_specifications):
    config = configparser.ConfigParser()
    # 'align' is missing and 'page-size' is incorrect
    config["Parameters"] = {
        "do-stuff": "no",
        "highlight": "yes",
        "organize": "yes",
        "page-width": 22,
    }
    expected_config_parameters = {
        "align": "center",
        "do_stuff": "no",
        "highlight": "yes",
        "organize": "yes",
        "page_width": 20,
        "paging_sizes": "10,25,100,All",
    }
    with mock.patch(f"{MODULE_PATH}.read_file_config", return_value=config):
        config_parameters = get_file_config(
            filename="does_not_matter",
            parameter_specifications=simple_parameter_specifications,
            directory=Path("/does_not_matter"),
        )
        assert config_parameters == expected_config_parameters


def test_get_file_config__incorrect_specification(
    caplog, simple_parameter_specifications
):
    caplog.set_level(logging.ERROR)
    config = configparser.ConfigParser()
    config["unknown"] = {"organize": "yes"}
    with mock.patch(f"{MODULE_PATH}.read_file_config", return_value=config):
        with pytest.raises(SystemExit):
            get_file_config(
                filename="does_not_matter",
                parameter_specifications=simple_parameter_specifications,
                directory=Path("/does_not_matter"),
            )
        assert "Parameters section not found in" in caplog.text


def test_get_external_links_from_config__correct_specification():
    config = configparser.ConfigParser()
    config["Links"] = {
        "altair": "\na\nb",
        "bokeh": "\nc\nd\n",
        "datatables": "\ne\n",
        "highlight_js": "\nf\ng",
        "jquery": "\nh",
        "plotly": "\ni",
    }
    expected_result = {
        "altair": ["a", "b"],
        "bokeh": ["c", "d"],
        "datatables": ["e"],
        "highlight_js": ["f", "g"],
        "jquery": ["h"],
        "plotly": ["i"],
    }
    with mock.patch(f"{MODULE_PATH}.read_file_config", return_value=config):
        result = get_external_links_from_config(
            filename="does_not_matter",
            directory=Path("/does_not_matter"),
        )
        assert result == expected_result


def test_get_external_links_from_config__incorrect_section(
    caplog, simple_parameter_specifications
):
    caplog.set_level(logging.ERROR)
    config = configparser.ConfigParser()
    # Wrong section name
    config["Unsupported"] = {
        "altair": "\na\nb",
        "bokeh": "\nc\nd\n",
        "datatables": "\ne\n",
        "highlight_js": "\nf\ng",
        "jquery": "\nh",
        "plotly": "\ni",
    }
    with mock.patch(f"{MODULE_PATH}.read_file_config", return_value=config):
        with pytest.raises(SystemExit):
            get_external_links_from_config(
                filename="does_not_matter",
                directory=Path("/does_not_matter"),
            )
        assert "section not found in" in caplog.text


def test_get_external_links_from_config__incorrect_keys(
    caplog, simple_parameter_specifications
):
    caplog.set_level(logging.ERROR)
    config = configparser.ConfigParser()
    # contains only some items
    config["Links"] = {
        "altair": "\na\nb",
        "bokeh": "\nc\nd\n",
    }
    with mock.patch(f"{MODULE_PATH}.read_file_config", return_value=config):
        with pytest.raises(SystemExit):
            get_external_links_from_config(
                filename="does_not_matter",
                directory=Path("/does_not_matter"),
            )
        assert "Configuration with items must contain links" in caplog.text


@pytest.mark.parametrize(
    "test_input_1,test_input_2,expected_result",
    [
        (None, None, None),
        ("a", None, "a"),
        (None, "b", "b"),
        ("a", "b", "a"),
    ],
)
def test_merge_values(test_input_1, test_input_2, expected_result):
    assert merge_values(test_input_1, test_input_2) == expected_result


@pytest.mark.parametrize(
    "primary_parameters,secondary_parameters,expected_result",
    [
        (
            {
                "align": None,
                "do_stuff": "no",
                "highlight": "yes",
                "organize": None,
                "page_width": 10,
                "paging_sizes": None,
            },
            {
                "align": "center",
                "do_stuff": None,
                "highlight": "no",
                "organize": None,
                "page_width": 20,
                "paging_sizes": "10,25,100,All",
            },
            {
                "align": "center",
                "do_stuff": "no",
                "highlight": "yes",
                "organize": None,
                "page_width": 10,
                "paging_sizes": "10,25,100,All",
            },
        ),
    ],
)
def test_merge_parameter_dictionaries(
    primary_parameters,
    secondary_parameters,
    expected_result,
    simple_parameter_specifications,
):
    assert (
        merge_parameter_dictionaries(
            primary_parameters,
            secondary_parameters,
            parameter_specifications=simple_parameter_specifications,
        )
        == expected_result
    )


@pytest.mark.parametrize(
    "test_input,expected_result",
    [
        ("None", None),
        ("yes", True),
        ("no", False),
        ("anything", "anything"),
        (42, 42),
    ],
)
def test_map_env_value(test_input, expected_result):
    assert _map_env_value(test_input) == expected_result


def test_get_parameter_value():
    pars = '{"a": 2, "html_dir_path": "/tmp/dir"}'
    with mock.patch.dict(os.environ, {"_TMP_PYREBALL_GENERATOR_PARAMETERS": pars}):
        assert get_parameter_value("a") == 2
        assert get_parameter_value("html_dir_path") == "/tmp/dir"
        assert get_parameter_value("html_dir_name") == "dir"
        assert get_parameter_value("html_file_path") == "/tmp/dir.html"


def test_make_sure_dir_exists(tmpdir):
    directory = str(tmpdir / "mydir")
    assert not os.path.exists(directory)
    make_sure_dir_exists(directory)
    assert os.path.exists(directory)
    # when executed again, nothing breaks
    make_sure_dir_exists(directory)
    assert os.path.exists(directory)


@mock.patch(f"{MODULE_PATH}.shutil")
def test_carefully_remove_directory_if_exists__doesnt_exist(shutil_mock, tmpdir):
    directory = Path(tmpdir / "whatever")
    assert not directory.exists()
    carefully_remove_directory_if_exists(directory)
    shutil_mock.rmtree.assert_not_called()


@mock.patch(f"{MODULE_PATH}.shutil")
def test_carefully_remove_directory_if_exists__error_when_deleting(shutil_mock, tmpdir):
    shutil_mock.rmtree.side_effect = OSError
    directory = Path(tmpdir / "whatever")
    directory.mkdir(parents=True)
    assert directory.exists()
    with pytest.raises(OSError):
        carefully_remove_directory_if_exists(directory)


def test_carefully_remove_directory_if_exists__empty_dir(tmpdir):
    directory = Path(tmpdir / "whatever")
    directory.mkdir(parents=True)
    assert directory.exists()
    carefully_remove_directory_if_exists(directory)
    assert not directory.exists()


def test_carefully_remove_directory_if_exists__with_image_files(tmpdir):
    directory = Path(tmpdir / "whatever")
    directory.mkdir(parents=True)

    # Creates empty files
    filenames = ["img.png", "img.jpg", "img.svg"]
    for filename in filenames:
        with open(directory / filename, "w") as fp:
            pass

    assert directory.exists()
    assert len(list(directory.glob("*"))) == len(filenames)
    carefully_remove_directory_if_exists(directory)
    assert not directory.exists()


def test_carefully_remove_directory_if_exists__with_non_image_files(tmpdir):
    directory = Path(tmpdir / "whatever")
    directory.mkdir(parents=True)

    # Creates empty files
    filenames = ["img.png", "img.jpg", "img.svg", "important_script.py"]
    for filename in filenames:
        with open(directory / filename, "w") as fp:
            pass

    assert directory.exists()
    assert len(list(directory.glob("*"))) == len(filenames)
    with pytest.raises(ValueError) as excinfo:
        carefully_remove_directory_if_exists(directory)
    assert "Cannot delete the original html directory" in str(excinfo.value)
    assert directory.exists()


@pytest.mark.parametrize(
    "text,repls,expected_result",
    [
        ("", [("apple", "pear"), ("dark", "light"), ("red", "green")], ""),
        (
            "This apple is dark red",
            [("apple", "pear"), ("dark", "light"), ("red", "green")],
            "This pear is light green",
        ),
        (
            "This apple is dark red",
            [
                ("apple", "pear"),
                ("dark", "light"),
                ("red", "green"),
                ("pear", "orange"),
            ],
            "This orange is light green",
        ),
        (
            "my aa and also aab ccdd cc table-5035059472 and also table-5035059472-6",
            [(r"aab?", "XX"), (r"cc(dd)?", "YY"), ("table-5035059472(-6)?", "table-6")],
            "my XX and also XX YY YY table-6 and also table-6",
        ),
    ],
)
def test_substitutor(text, repls, expected_result):
    substitutor = Substitutor(repls)
    result = substitutor.sub(text)
    assert result == expected_result



from pathlib import Path

import pkg_resources

CONFIG_INI_FILENAME = "config.ini"
LINKS_INI_FILENAME = "external_links.ini"
STYLES_TEMPLATE_FILENAME = "css.template"
HTML_TEMPLATE_FILENAME = "html.template"
DEFAULT_PATH_TO_CONFIG = Path(pkg_resources.resource_filename("pyreball", "cfg"))

PILCROW_SIGN = "¶"
NON_BREAKABLE_SPACE = "\u00A0"



import logging
from pathlib import Path

import pytest

from pyreball.utils.template import get_css, get_html


def test_get_html(tmpdir):
    filename = "styles"
    template_path = Path(tmpdir) / filename
    with open(template_path, "w") as f:
        f.write(
            "<html>title: <!--PYREBALL_PAGE_TITLE-->, "
            "css: <!--PYREBALL_CSS_DEFINITIONS-->"
            "<!--PYREBALL_REPORT_CONTENTS-->"
            "</html>"
        )
    result_begin, result_end = get_html(
        template_path=template_path,
        title="t1",
        css_definitions="c1",
    )
    assert result_begin == "<html>title: t1, css: c1"
    assert result_end == "</html>"


def test_get_css__existing_file(tmpdir):
    filename = "styles"
    directory = Path(tmpdir)
    with open(directory / filename, "w") as f:
        f.write("body {width: {{page_width}}%;}\n")
    styles = get_css(filename, directory, page_width=30)
    assert styles == "body {width: 30%;}\n"


def test_get_css__non_existing_file(tmpdir, caplog):
    caplog.set_level(logging.ERROR)
    filename = "styles"
    directory = Path(tmpdir)
    with pytest.raises(SystemExit):
        get_css(filename, directory, page_width=30)
    assert "There was a problem reading file" in caplog.text



"""Text utils for creating strings with HTML elements."""
from typing import Any, List, Literal, Optional

from pyreball._common import AttrsParameter, ClParameter


def _construct_attrs_str(attrs: AttrsParameter) -> Optional[str]:
    if not attrs:
        return None
    return " ".join(
        [f'{k}="{v}"' if v is not None else f"{k}" for k, v in attrs.items()]
    )


def _construct_class_attr_string(cl: ClParameter) -> Optional[str]:
    if cl is None:
        return None
    if isinstance(cl, str):
        cl_values = cl
    else:
        cl_values = " ".join(cl)
    return _construct_attrs_str({"class": cl_values})


def bold(
    *values: Any, cl: ClParameter = None, attrs: AttrsParameter = None, sep: str = ""
) -> str:
    """
    Create a `<bold>` element string with given values.

    Args:
        *values: Zero or more values to be enclosed in the tag.
            All values are converted to strings.
        cl: One or more class names to be added to the tag.
            If a string is provided, it is used as it is.
            If a list of strings is provided, the strings are joined with a space.
            If `None`, no class is added.
            If an empty list is provided, class attribute is added with an empty string.
        attrs: Additional attributes to be added to the tag.
            Dictionary `{"key1": "value1", ..., "keyN": "valueN"}`
            is converted to `key1="value1" ... keyN="valueN"`.
            To construct boolean HTML attributes,
            set `None` for given key. Any quotes in values are not escaped.
        sep: String separator of the values. Defaults to an empty string.

    Returns:
        HTML string representing the tag with given values.
    """
    return tag(*values, name="b", cl=cl, attrs=attrs, sep=sep)


def em(
    *values: Any, cl: ClParameter = None, attrs: AttrsParameter = None, sep: str = ""
) -> str:
    """
    Create a `<em>` element string with given values.

    Args:
        *values: Zero or more values to be enclosed in the tag.
            All values are converted to strings.
        cl: One or more class names to be added to the tag.
            If a string is provided, it is used as it is.
            If a list of strings is provided, the strings are joined with a space.
            If `None`, no class is added.
            If an empty list is provided, class attribute is added with an empty string.
        attrs: Additional attributes to be added to the tag.
            Dictionary `{"key1": "value1", ..., "keyN": "valueN"}`
            is converted to `key1="value1" ... keyN="valueN"`.
            To construct boolean HTML attributes,
            set `None` for given key. Any quotes in values are not escaped.
        sep: String separator of the values. Defaults to an empty string.

    Returns:
        HTML string representing the tag with given values.
    """
    return tag(*values, name="em", cl=cl, attrs=attrs, sep=sep)


def _collect_classes_for_code_strings(
    initial_class_list: List[str],
    cl: ClParameter,
    syntax_highlight: Optional[Literal["python"]],
) -> ClParameter:
    classes_to_be_added = initial_class_list[:]
    if syntax_highlight is not None:
        if syntax_highlight == "python":
            classes_to_be_added.append("python")
        else:
            raise ValueError(
                f"Unsupported syntax highlighting language: {syntax_highlight}."
            )
        if cl is None:
            cl = classes_to_be_added
        elif isinstance(cl, str):
            cl = [cl] + classes_to_be_added
        else:
            cl = cl + classes_to_be_added
    return cl


def code(
    *values: Any,
    cl: ClParameter = None,
    attrs: AttrsParameter = None,
    sep: str = "",
    syntax_highlight: Optional[Literal["python"]] = "python",
) -> str:
    """
    Create a `<code>` element string with given values.

    This element is used to display a source code inline.
    It is possible to highlight the code syntax by setting `syntax_highlight`
    parameter to an appropriate string.

    Args:
        *values: Zero or more values to be enclosed in the tag.
            All values are converted to strings.
        cl: One or more class names to be added to the tag.
            If a string is provided, it is used as it is.
            If a list of strings is provided, the strings are joined with a space.
            If `None`, no class is added.
            If an empty list is provided, class attribute is added with an empty string.
        attrs: Additional attributes to be added to the tag.
            Dictionary `{"key1": "value1", ..., "keyN": "valueN"}`
            is converted to `key1="value1" ... keyN="valueN"`.
            To construct boolean HTML attributes,
            set `None` for given key. Any quotes in values are not escaped.
        sep: String separator of the values. Defaults to an empty string.
        syntax_highlight: Syntax highlighting language.
            Currently only `'python'` is supported. If `None`, no highlight is applied.
            When highlight is turned on, language name and `'inline-highlight'`
            are added to the element as classes.

    Returns:
        HTML string representing the tag with given values.
    """
    cl = _collect_classes_for_code_strings(["inline-highlight"], cl, syntax_highlight)
    return tag(*values, name="code", cl=cl, attrs=attrs, sep=sep)


def code_block(
    *values: Any,
    cl: ClParameter = None,
    attrs: AttrsParameter = None,
    pre_cl: ClParameter = None,
    pre_attrs: AttrsParameter = None,
    sep: str = "",
    syntax_highlight: Optional[Literal["python"]] = "python",
) -> str:
    """
    Create a `<pre><code>` pair element string with given values.

    This element is used to display a source code in a block.
    It is possible to highlight the code syntax by setting `syntax_highlight`
    parameter to an appropriate string.

    Args:
        *values: Zero or more values to be enclosed in the tag.
            All values are converted to strings.
        cl: One or more class names to be added to the <code> tag.
            If a string is provided, it is used as it is.
            If a list of strings is provided, the strings are joined with a space.
            If `None`, no class is added.
            If an empty list is provided, class attribute is added with an empty string.
            A class is also added when `syntax_highlight` is set.
        attrs: Additional attributes to be added to the <code> tag.
            Dictionary `{"key1": "value1", ..., "keyN": "valueN"}`
            is converted to `key1="value1" ... keyN="valueN"`.
            To construct boolean HTML attributes,
            set `None` for given key. Any quotes in values are not escaped.
        pre_cl: The same as `cl` parameter, but for the `<pre>` tag.
        pre_attrs: The same as `attrs` parameter, but for the `<pre>` tag.
        sep: String separator of the values. Defaults to an empty string.
        syntax_highlight: Syntax highlighting language.
            Currently only `'python'` is supported. If `None`, no highlight is applied.
            When highlight is turned on, language name is added to the `<code>`
            element as a class.

    Returns:
        HTML string representing the tag with given values.
    """
    cl = _collect_classes_for_code_strings([], cl, syntax_highlight)
    code_text = tag(*values, name="code", cl=cl, attrs=attrs, sep=sep)
    return tag(code_text, name="pre", cl=pre_cl, attrs=pre_attrs)


def div(
    *values: Any, cl: ClParameter = None, attrs: AttrsParameter = None, sep: str = ""
) -> str:
    """
    Create a `<div>` element string with given values.

    Args:
        *values: Zero or more values to be enclosed in the tag.
            All values are converted to strings.
        cl: One or more class names to be added to the tag.
            If a string is provided, it is used as it is.
            If a list of strings is provided, the strings are joined with a space.
            If `None`, no class is added.
            If an empty list is provided, class attribute is added with an empty string.
        attrs: Additional attributes to be added to the tag.
            Dictionary `{"key1": "value1", ..., "keyN": "valueN"}`
            is converted to `key1="value1" ... keyN="valueN"`.
            To construct boolean HTML attributes,
            set `None` for given key. Any quotes in values are not escaped.
        sep: String separator of the values. Defaults to an empty string.

    Returns:
        HTML string representing the tag with given values.
    """
    return tag(*values, name="div", cl=cl, attrs=attrs, sep=sep)


def span(
    *values: Any, cl: ClParameter = None, attrs: AttrsParameter = None, sep: str = ""
) -> str:
    """
    Create a `<span>` element string with given values.

    Args:
        *values: Zero or more values to be enclosed in the tag.
            All values are converted to strings.
        cl: One or more class names to be added to the tag.
            If a string is provided, it is used as it is.
            If a list of strings is provided, the strings are joined with a space.
            If `None`, no class is added.
            If an empty list is provided, class attribute is added with an empty string.
        attrs: Additional attributes to be added to the tag.
            Dictionary `{"key1": "value1", ..., "keyN": "valueN"}`
            is converted to `key1="value1" ... keyN="valueN"`.
            To construct boolean HTML attributes,
            set `None` for given key. Any quotes in values are not escaped.
        sep: String separator of the values. Defaults to an empty string.

    Returns:
        HTML string representing the tag with given values.
    """
    return tag(*values, name="span", cl=cl, attrs=attrs, sep=sep)


def a(
    *values: Any, cl: ClParameter = None, attrs: AttrsParameter = None, sep: str = ""
) -> str:
    """
    Create a `<a>` element string with given values.

    Args:
        *values: Zero or more values to be enclosed in the tag.
            All values are converted to strings.
        cl: One or more class names to be added to the tag.
            If a string is provided, it is used as it is.
            If a list of strings is provided, the strings are joined with a space.
            If `None`, no class is added.
            If an empty list is provided, class attribute is added with an empty string.
        attrs: Additional attributes to be added to the tag.
            Dictionary `{"key1": "value1", ..., "keyN": "valueN"}`
            is converted to `key1="value1" ... keyN="valueN"`.
            To construct boolean HTML attributes,
            set `None` for given key. Any quotes in values are not escaped.
        sep: String separator of the values. Defaults to an empty string.

    Returns:
        HTML string representing the tag with given values.
    """
    return tag(*values, name="a", cl=cl, attrs=attrs, sep=sep)


def link(text: str, href: str) -> str:
    """
    Create a `<a>` element string with given href and text.
    This function is a shortcut for `a()` function.

    Args:
        text: Text of the link.
        href: URL of the link.

    Returns:
        HTML string representing the tag with given values.
    """
    return a(text, attrs={"href": href})


def _enclose_into_li_tags(
    *values: Any,
    cl: ClParameter = None,
    attrs: AttrsParameter = None,
    sep: str = "",
):
    result = ""
    for value in values:
        if isinstance(value, tuple):
            value_str = "".join(map(str, value))
        else:
            value_str = str(value)
        result += tag(value_str, name="li", cl=cl, attrs=attrs, sep=sep)
    return result


def ulist(
    *values: Any,
    cl: ClParameter = None,
    attrs: AttrsParameter = None,
    li_cl: ClParameter = None,
    li_attrs: AttrsParameter = None,
    sep: str = "",
) -> str:
    """
    Create a `<ul>` element string with values being enclosed into `<li>` tags.

    Args:
        *values: Zero or more values to be enclosed in the tag.
            All values are converted to strings.
        cl: One or more class names to be added to the `<ul>` tag.
            If a string is provided, it is used as it is.
            If a list of strings is provided, the strings are joined with a space.
            If `None`, no class is added.
            If an empty list is provided, class attribute is added with an empty string.
        attrs: Additional attributes to be added to the `<ul>` tag.
            Dictionary `{"key1": "value1", ..., "keyN": "valueN"}`
            is converted to `key1="value1" ... keyN="valueN"`.
            To construct boolean HTML attributes,
            set `None` for given key. Any quotes in values are not escaped.
        li_cl: The same as `cl` parameter, but for the `<li>` tag.
        li_attrs: The same as `attrs` parameter, but for the `<li>` tag.
        sep: Separator string to concatenate the values with.

    Returns:
        HTML string representing the tag with given values.
    """
    inner_content = _enclose_into_li_tags(*values, cl=li_cl, attrs=li_attrs, sep=sep)
    return tag(inner_content, name="ul", cl=cl, attrs=attrs, sep=sep)


def olist(
    *values: Any,
    cl: ClParameter = None,
    attrs: AttrsParameter = None,
    li_cl: ClParameter = None,
    li_attrs: AttrsParameter = None,
    sep: str = "",
) -> str:
    """
    Create a `<ol>` element string with values being enclosed into `<li>` tags.

    Args:
        *values: Zero or more values to be enclosed in the tag.
            All values are converted to strings.
        cl: One or more class names to be added to the `<ol>` tag.
            If a string is provided, it is used as it is.
            If a list of strings is provided, the strings are joined with a space.
            If `None`, no class is added.
            If an empty list is provided, class attribute is added with an empty string.
        attrs: Additional attributes to be added to the `<ol>` tag.
            Dictionary `{"key1": "value1", ..., "keyN": "valueN"}`
            is converted to `key1="value1" ... keyN="valueN"`.
            To construct boolean HTML attributes,
            set `None` for given key. Any quotes in values are not escaped.
        li_cl: The same as `cl` parameter, but for the `<li>` tag.
        li_attrs: The same as `attrs` parameter, but for the `<li>` tag.
        sep: Separator string to concatenate the values with.

    Returns:
        HTML string representing the tag with given values.
    """
    inner_content = _enclose_into_li_tags(*values, cl=li_cl, attrs=li_attrs, sep=sep)
    return tag(inner_content, name="ol", cl=cl, attrs=attrs, sep=sep)


def tag(
    *values: Any,
    name: str,
    cl: ClParameter = None,
    attrs: AttrsParameter = None,
    sep: str = "",
    paired: bool = True,
) -> str:
    """
    Create a tag string with given values.

    Args:

        *values: Zero or more values to be enclosed in the tag.
            All values are converted to strings.
        name: Name of the tag.
        cl: One or more class names to be added to the tag.
            If a string is provided, it is used as it is.
            If a list of strings is provided, the strings are joined with a space.
            If `None`, no class is added.
            If an empty list is provided, class attribute is added with an empty string.
        attrs: Additional attributes to be added to the tag.
            Dictionary `{"key1": "value1", ..., "keyN": "valueN"}`
            is converted to `key1="value1" ... keyN="valueN"`.
            To construct boolean HTML attributes,
            set `None` for given key. Any quotes in values are not escaped.
        sep: String separator of the values. Defaults to an empty string.
        paired: If True, the values are enclosed in a pair of tags.
            Otherwise, unpaired tag is created.

    Returns:
        HTML string representing the tag with given values.
    """
    cl_str = _construct_class_attr_string(cl)
    if cl_str and attrs is not None and "class" in attrs:
        raise ValueError(
            "class attribute cannot be set through cl and attrs parameters "
            "at the same time!"
        )
    attrs_str = _construct_attrs_str(attrs)
    opening_tag = (
        f"{name}{f' {cl_str}' if cl_str else ''}{f' {attrs_str}' if attrs_str else ''}"
    )
    if paired:
        inner_contents = sep.join(map(str, values))
        return f"<{opening_tag}>{sep}{inner_contents}{sep}</{name}>"
    else:
        return f"<{opening_tag}>"



__version__ = "1.0.0"

from pyreball.html import (
    print,
    print_code_block,
    print_div,
    print_figure,
    print_h1,
    print_h2,
    print_h3,
    print_h4,
    print_h5,
    print_h6,
    print_table,
    Reference,
    set_title,
)
from pyreball.text import (
    a,
    bold,
    code,
    code_block,
    div,
    em,
    link,
    olist,
    span,
    tag,
    ulist,
)

__all__ = [
    "print_figure",
    "print",
    "print_code_block",
    "print_div",
    "print_h1",
    "print_h2",
    "print_h3",
    "print_h4",
    "print_h5",
    "print_h6",
    "print_table",
    "Reference",
    "set_title",
    "a",
    "bold",
    "code",
    "code_block",
    "div",
    "em",
    "link",
    "olist",
    "span",
    "tag",
    "ulist",
]



from copy import copy

import pytest

from pyreball.text import (
    _collect_classes_for_code_strings,
    _construct_attrs_str,
    _construct_class_attr_string,
    a,
    bold,
    code,
    code_block,
    div,
    em,
    link,
    olist,
    span,
    tag,
    ulist,
)


@pytest.mark.parametrize(
    "test_input,expected_result",
    [
        (None, None),
        ("", 'class=""'),
        ([], 'class=""'),
        ("ab", 'class="ab"'),
        ("ab xy", 'class="ab xy"'),
        (["ab"], 'class="ab"'),
        (["ab", "e", "fg"], 'class="ab e fg"'),
    ],
)
def test__construct_class_attr_string(test_input, expected_result):
    assert _construct_class_attr_string(test_input) == expected_result


@pytest.mark.parametrize(
    "test_input,expected_result",
    [
        (
            None,
            None,
        ),
        (
            {},
            None,
        ),
        (
            {"x": "y"},
            'x="y"',
        ),
        (
            {"x": "y", "checked": None, "a": "b c", "valid": None},
            'x="y" checked a="b c" valid',
        ),
    ],
)
def test__construct_attrs_str(test_input, expected_result):
    assert _construct_attrs_str(test_input) == expected_result


def test_bold():
    assert bold("a", "b", sep="\n") == "<b>\na\nb\n</b>"


def test_em():
    assert em("a", "b", sep="\n") == "<em>\na\nb\n</em>"


@pytest.mark.parametrize(
    "initial_class_list,cl,syntax_highlight,expected_result",
    [
        ([], None, None, None),
        ([], [], None, []),
        ([], "cl1", None, "cl1"),
        ([], ["cl1"], None, ["cl1"]),
        ([], None, "python", ["python"]),
        ([], [], "python", ["python"]),
        ([], "cl1", "python", ["cl1", "python"]),
        ([], ["cl1"], "python", ["cl1", "python"]),
        (["x"], None, None, None),
        (["x"], [], None, []),
        (["x"], "cl1", None, "cl1"),
        (["x"], ["cl1"], None, ["cl1"]),
        (["x"], None, "python", ["x", "python"]),
        (["x"], [], "python", ["x", "python"]),
        (["x"], "cl1", "python", ["cl1", "x", "python"]),
        (["x"], ["cl1"], "python", ["cl1", "x", "python"]),
    ],
)
def test__collect_classes_for_code_strings(
    initial_class_list, cl, syntax_highlight, expected_result
):
    original_cl = copy(cl)
    original_initial_class_list = copy(initial_class_list)
    original_syntax_highlight = copy(syntax_highlight)
    result = _collect_classes_for_code_strings(initial_class_list, cl, syntax_highlight)
    assert result == expected_result
    # check that input values haven't changed
    assert original_cl == cl
    assert original_initial_class_list == initial_class_list
    assert original_syntax_highlight == syntax_highlight


def test_code__without_syntax_highlight():
    expected_result = "<code>\na\nb\n</code>"
    assert code("a", "b", sep="\n", syntax_highlight=None) == expected_result


def test_code__with_syntax_highlight():
    expected_result = '<code class="inline-highlight python">\na\nb\n</code>'
    assert code("a", "b", sep="\n", syntax_highlight="python") == expected_result


def test_code__unsupported_syntax_highlight():
    with pytest.raises(ValueError):
        code("a", "b", sep="\n", syntax_highlight="my_new_lang")


def test_code_block__without_syntax_highlight():
    expected_result = "<pre><code>\na\nb\n</code></pre>"
    assert code_block("a", "b", sep="\n", syntax_highlight=None) == expected_result


def test_code_block__with_syntax_highlight():
    expected_result = '<pre><code class="python">\na\nb\n</code></pre>'
    assert code_block("a", "b", sep="\n", syntax_highlight="python") == expected_result


def test_code_block__with_syntax_highlight_and_attributes():
    expected_result = (
        '<pre class="pre1" pa="pv">'
        '<code class="code1 python" ca="cv">\na\nb\n</code>'
        "</pre>"
    )
    result = code_block(
        "a",
        "b",
        cl="code1",
        attrs={"ca": "cv"},
        pre_cl="pre1",
        pre_attrs={"pa": "pv"},
        sep="\n",
        syntax_highlight="python",
    )
    assert result == expected_result


def test_code_block__unsupported_syntax_highlight():
    with pytest.raises(ValueError):
        code_block("a", "b", sep="\n", syntax_highlight="my_new_lang")


def test_div():
    assert div("a", "b", sep="\n") == "<div>\na\nb\n</div>"


def test_span():
    assert span("a", "b", sep="\n") == "<span>\na\nb\n</span>"


def test_a():
    expected_result = '<a href="www.example.com">\na\nb\n</a>'
    assert a("a", "b", attrs={"href": "www.example.com"}, sep="\n") == expected_result


def test_link():
    assert link("my text", "myurl") == '<a href="myurl">my text</a>'


@pytest.mark.parametrize(
    "test_input,expected_result",
    [
        (
            ["a", "b", "c"],
            "<ul><li>a</li><li>b</li><li>c</li></ul>",
        ),
        (
            ["a", 43, 42],
            "<ul><li>a</li><li>43</li><li>42</li></ul>",
        ),
        (
            [("a", "<ul><li>b</li></ul>")],
            "<ul><li>a<ul><li>b</li></ul></li></ul>",
        ),
        (
            [("b", "<ol><li>x</li></ol>"), ("a", "<ul><li>y</li><li>z</li></ul>")],
            "<ul><li>b<ol><li>x</li></ol></li><li>a<ul><li>y</li><li>z</li></ul></li></ul>",
        ),
    ],
)
def test_ulist__without_parameters(test_input, expected_result):
    assert ulist(*test_input) == expected_result


def test_ulist__with_parameters():
    input_kwargs = {
        "cl": "ul_class",
        "attrs": {"x": "y"},
        "li_cl": "li_class",
        "li_attrs": {"a": "b"},
    }
    test_input = ["value1", "value2", "value3"]
    expected_result = (
        '<ul class="ul_class" x="y">'
        '<li class="li_class" a="b">value1</li>'
        '<li class="li_class" a="b">value2</li>'
        '<li class="li_class" a="b">value3</li>'
        "</ul>"
    )
    assert ulist(*test_input, **input_kwargs) == expected_result


@pytest.mark.parametrize(
    "test_input,expected_result",
    [
        (
            ["a", "b", "c"],
            "<ol><li>a</li><li>b</li><li>c</li></ol>",
        ),
        (
            ["a", 43, 42],
            "<ol><li>a</li><li>43</li><li>42</li></ol>",
        ),
        (
            [("a", "<ol><li>b</li></ol>")],
            "<ol><li>a<ol><li>b</li></ol></li></ol>",
        ),
        (
            [("b", "<ol><li>x</li></ol>"), ("a", "<ol><li>y</li><li>z</li></ol>")],
            "<ol><li>b<ol><li>x</li></ol></li><li>a<ol><li>y</li><li>z</li></ol></li></ol>",
        ),
    ],
)
def test_olist__without_parameters(test_input, expected_result):
    assert olist(*test_input) == expected_result


def test_olist__with_parameters():
    input_kwargs = {
        "cl": "ul_class",
        "attrs": {"x": "y"},
        "li_cl": "li_class",
        "li_attrs": {"a": "b"},
    }
    test_input = ["value1", "value2", "value3"]
    expected_result = (
        '<ol class="ul_class" x="y">'
        '<li class="li_class" a="b">value1</li>'
        '<li class="li_class" a="b">value2</li>'
        '<li class="li_class" a="b">value3</li>'
        "</ol>"
    )
    assert olist(*test_input, **input_kwargs) == expected_result


@pytest.mark.parametrize(
    "input_args,input_kwargs,expected_result",
    [
        (
            ["abc"],
            {"name": "x", "cl": None, "attrs": None, "sep": ""},
            "<x>abc</x>",
        ),
        (
            [2351],
            {"name": "x", "cl": None, "attrs": None, "sep": ""},
            "<x>2351</x>",
        ),
        (
            [2351, "hello"],
            {"name": "abc", "cl": None, "attrs": None, "sep": "\n"},
            "<abc>\n2351\nhello\n</abc>",
        ),
        (
            [2351, "hello"],
            {"name": "abc", "cl": "cl1", "attrs": None, "sep": "\n"},
            '<abc class="cl1">\n2351\nhello\n</abc>',
        ),
        (
            [2351, "hello"],
            {"name": "abc", "cl": "cl1 cl2", "attrs": None, "sep": "\n"},
            '<abc class="cl1 cl2">\n2351\nhello\n</abc>',
        ),
        (
            [2351, "hello"],
            {"name": "abc", "cl": ["cl1", "cl2"], "attrs": None, "sep": "\n"},
            '<abc class="cl1 cl2">\n2351\nhello\n</abc>',
        ),
        (
            [2351, "hello"],
            {"name": "abc", "cl": None, "attrs": {"key": "val"}, "sep": "\n"},
            '<abc key="val">\n2351\nhello\n</abc>',
        ),
        (
            [2351, "hello"],
            {"name": "abc", "cl": ["cl1", "cl2"], "attrs": {"key": "val"}, "sep": "\n"},
            '<abc class="cl1 cl2" key="val">\n2351\nhello\n</abc>',
        ),
        (
            [],
            {
                "name": "abc",
                "cl": ["cl1", "cl2"],
                "attrs": {"key": "val"},
                "paired": False,
            },
            '<abc class="cl1 cl2" key="val">',
        ),
    ],
)
def test_tag(input_args, input_kwargs, expected_result):
    assert tag(*input_args, **input_kwargs) == expected_result


def test_tag__duplicate_class_setting():
    with pytest.raises(ValueError):
        tag("value", name="x", cl="my_class", attrs={"class": "my_cl"}, sep="")



"""Main functions that serve as building blocks of the final html file."""
import builtins
import io
import json
import os
import random
import re
from typing import (
    Any,
    cast,
    Dict,
    List,
    Literal,
    Optional,
    Set,
    Tuple,
    TYPE_CHECKING,
    Union,
)

from pyreball._common import AttrsParameter, ClParameter
from pyreball.constants import NON_BREAKABLE_SPACE, PILCROW_SIGN

from pyreball.text import code_block, div
from pyreball.utils.param import get_parameter_value, make_sure_dir_exists, merge_values

if TYPE_CHECKING:
    # needed for mypy
    # noinspection PyUnresolvedReferences,PyPackageRequirements
    import altair  # type: ignore

    # noinspection PyPackageRequirements
    import bokeh  # type: ignore

    # noinspection PyPackageRequirements
    import matplotlib  # type: ignore

    # noinspection PyPackageRequirements
    import pandas  # type: ignore

    # noinspection PyPackageRequirements
    import plotly  # type: ignore

AltairFigType = Union[
    "altair.vegalite.v4.api.Chart",
    "altair.vegalite.v4.api.ConcatChart",
    "altair.vegalite.v4.api.FacetChart",
    "altair.vegalite.v4.api.HConcatChart",
    "altair.vegalite.v4.api.LayerChart",
    "altair.vegalite.v4.api.RepeatChart",
    "altair.vegalite.v4.api.VConcatChart",
]
FigType = Union[
    "matplotlib.figure.Figure",
    "plotly.graph_objs._figure.Figure",
    "altair.vegalite.v4.api.Chart",
    "altair.vegalite.v4.api.ConcatChart",
    "altair.vegalite.v4.api.FacetChart",
    "altair.vegalite.v4.api.HConcatChart",
    "altair.vegalite.v4.api.LayerChart",
    "altair.vegalite.v4.api.RepeatChart",
    "altair.vegalite.v4.api.VConcatChart",
]

_references: Set[str] = set()
_heading_memory: Dict[str, Any] = {}
_code_block_memory: Dict[str, Any] = {}
_table_memory: Dict[str, Any] = {}
_graph_memory: Dict[str, Any] = {}

ALIGN_CLASS_MAP = {
    "center": "pyreball-centered",
    "left": "pyreball-left-aligned",
    "right": "pyreball-right-aligned",
}


class Reference:
    """
    Class for creating references, i.e. anchors in HTML.
    """

    def __init__(self, default_text: Optional[str] = None) -> None:
        """
        Create a new reference.

        Args:
            default_text: Default text of the link.
                This text can be overriden by parameter of `__call__` method.
                If not provided, Pyreball automatically inserts a text.
                For tables, images and code blocks, their number is used.
                For headings, their text is used.
        """
        self.id = f"id{random.getrandbits(64)}"
        self.text = default_text

    def __str__(self) -> str:
        """
        Create a link string with default text.

        Returns:
            Link string.
        """
        return (
            f'<a href="#ref-{self.id}">'
            f"{self.id if self.text is None else self.text}"
            f"</a>"
        )

    def __call__(self, text: str) -> str:
        """
        Create a link string with given text.

        Args:
            text: Text of the link.

        Returns:
            Link string.
        """
        return f'<a href="#ref-{self.id}">{text}</a>'


def _check_and_mark_reference(reference: Reference) -> None:
    """Check and save a reference.

    This function is used when references are added to tables or figures.
    If a table or a figure is about to get a reference that was already used
    for another object, an error is raised.
    """
    if reference.id in _references:
        raise ValueError(
            "Reference is used for the second time. "
            "You have to create another reference for this object."
        )
    else:
        _references.add(reference.id)


def set_title(title: str) -> None:
    """
    Set page title.

    Note that this function does not have to be called at the beginning of the script.
    If this function is not called via pyreball and parameter keep_stdout
    is set to True, it just prints the title to stdout.

    Args:
        title: Title string.
    """
    if not get_parameter_value("html_file_path") or get_parameter_value("keep_stdout"):
        builtins.print(title)
    if get_parameter_value("html_file_path"):
        # it is assumed that the heading is already written into the file,
        # so find the line with title element and replace its contents
        with open(get_parameter_value("html_file_path"), "r") as f:
            lines = f.readlines()

        # replace the title and also add "custom_pyreball_title" class,
        # so that we know it was replaced by this function
        lines = [
            re.sub(
                r"^<title>[^<]*</title>",
                f'<title class="custom_pyreball_title">{title}</title>',
                line,
            )
            for line in lines
        ]
        with open(get_parameter_value("html_file_path"), "w") as f:
            f.writelines(lines)


def _write_to_html(string: str, end: str = "\n") -> None:
    if get_parameter_value("html_file_path"):
        with open(get_parameter_value("html_file_path"), "a") as f:
            f.write(string)
            f.write(end)


def _tidy_title(title: str) -> str:
    """
    Transforms title into lowercase alphanumerical sequence separated by underscores.

    Args:
        title: The string that you want to transform.

    Returns:
        str: The transformed string.
    """
    new_title = re.sub(
        "([a-z])([A-Z])", r"\g<1> \g<2>", title
    )  # CamelCase -> words with spaces
    new_title = new_title.lower()  # to lower case
    new_title = re.sub(r"%", "_percent", new_title)  # replace % symbol with "_percent"
    new_title = re.sub(
        r"[=-\\.,]", " ", new_title
    )  # replace some special characters with spaces
    new_title = re.sub(
        r"[^\w\s]", "", new_title
    ).strip()  # remove remaining special characters
    new_title = re.sub(
        r"\s+", "_", new_title
    )  # replace sequences of spaces with an underscore
    return new_title


def _reduce_whitespaces(string: str) -> str:
    string = re.sub(r"\s+", " ", string)
    return string.strip()


def _get_heading_number(level: int, l_heading_counting: List[int]) -> str:
    return ".".join(map(str, l_heading_counting[:level]))


def _print_heading(
    string: str, level: int = 1, reference: Optional[Reference] = None
) -> None:
    if level > 6:
        raise ValueError("Heading level cannot be greater than 6.")
    if level < 1:
        raise ValueError("Heading level cannot be less than 1.")

    if "heading_index" not in _heading_memory:
        _heading_memory["heading_index"] = 1

    heading_index = _heading_memory["heading_index"]

    if get_parameter_value("numbered_headings"):
        if "heading_counting" not in _heading_memory:
            # what is the index of current h1, h2, h3, h4, h5, h6?
            _heading_memory["heading_counting"] = [0, 0, 0, 0, 0, 0]

        # increase the number in the level
        _heading_memory["heading_counting"][level - 1] = (
            _heading_memory["heading_counting"][level - 1] + 1
        )
        # reset all sub-levels
        _heading_memory["heading_counting"][level:] = [0] * (6 - level)
        # get the string of the numbered section and append non-breakable space
        heading_number_str = (
            _get_heading_number(level, _heading_memory["heading_counting"])
            + NON_BREAKABLE_SPACE
            + NON_BREAKABLE_SPACE
        )
    else:
        heading_number_str = ""

    string = heading_number_str + _reduce_whitespaces(string)
    # use heading_index in the id of the heading,
    # so there are no collisions in the case of same texts
    if reference:
        _check_and_mark_reference(reference)
        tidy_string = f"ch_{reference.id}_{_tidy_title(string)}_{heading_index}"
    else:
        tidy_string = f"ch_{_tidy_title(string)}_{heading_index}"

    if not get_parameter_value("html_file_path") or get_parameter_value("keep_stdout"):
        builtins.print(string.replace(NON_BREAKABLE_SPACE * 2, " "))

    if get_parameter_value("html_file_path"):
        header_contents = (
            f"{string}"
            f'<a class="pyreball-anchor-link" href="#{tidy_string}">{PILCROW_SIGN}</a>'
        )
        # For correct functioning of references,
        # it is expected that single line contains at most one heading,
        # and the heading is whole there with all links.
        _write_to_html(f'<h{level} id="{tidy_string}">{header_contents}</h{level}>')
        _heading_memory["heading_index"] += 1


def print_h1(string: str, reference: Optional[Reference] = None) -> None:
    """
    Print h1 heading.

    Args:
        string: Content of the heading.
        reference: Reference object.
    """
    _print_heading(string, level=1, reference=reference)


def print_h2(string: str, reference: Optional[Reference] = None) -> None:
    """
    Print h2 heading.

    Args:
        string: Content of the heading.
        reference: Reference object.
    """
    _print_heading(string, level=2, reference=reference)


def print_h3(string: str, reference: Optional[Reference] = None) -> None:
    """
    Print h3 heading.

    Args:
        string: Content of the heading.
        reference: Reference object.
    """
    _print_heading(string, level=3, reference=reference)


def print_h4(string: str, reference: Optional[Reference] = None) -> None:
    """
    Print h4 heading.

    Args:
        string: Content of the heading.
        reference: Reference object.
    """
    _print_heading(string, level=4, reference=reference)


def print_h5(string: str, reference: Optional[Reference] = None) -> None:
    """
    Print h5 heading.

    Args:
        string: Content of the heading.
        reference: Reference object.
    """
    _print_heading(string, level=5, reference=reference)


def print_h6(string: str, reference: Optional[Reference] = None) -> None:
    """
    Print h6 heading.

    Args:
        string: Content of the heading.
        reference: Reference object.
    """
    _print_heading(string, level=6, reference=reference)


def print_div(
    *values: Any,
    cl: ClParameter = None,
    attrs: AttrsParameter = None,
    sep: str = "",
    end: str = "\n",
) -> None:
    """
    Print values into a div element.

    Any value that is not a string is converted to a string first.

    Args:
        *values: Zero or more values to be printed into the div.
        cl: One or more class names to be added to the `<div>` tag.
            If a string is provided, it is used as it is.
            If a list of strings is provided, the strings are joined with space.
            If `None`, no class is added.
            If an empty list is provided, class attribute is added with an empty string.
        attrs: Additional attributes to be added to the `<div>` tag.
            Dictionary `{"key1": "value1", ..., "keyN": "valueN"}`
            is converted to `key1="value1" ... keyN="valueN"`.
            To construct boolean HTML attributes, set `None` for given key.
            Any quotes in values are not escaped.
        sep: String separator of the values inside the tag. Defaults to an empty string.
        end: String appended after the tag. Defaults to a newline.
    """
    div_str = div(*values, cl=cl, attrs=attrs, sep=sep)
    print(div_str, end=end)


def _wrap_code_block_html(
    source_code_str: str,
    code_block_index: int = 0,
    caption: Optional[str] = None,
    reference: Optional[Reference] = None,
    align: str = "center",
    caption_position: str = "bottom",
    numbered: bool = True,
    sep: str = "",
):
    if reference:
        _check_and_mark_reference(reference)
        anchor_link = f"code-block-{reference.id}-{code_block_index}"
    else:
        anchor_link = f"code-block-{code_block_index}"

    caption_element = _prepare_caption_element(
        prefix="Source",
        caption=caption,
        numbered=numbered,
        index=code_block_index,
        anchor_link=anchor_link,
    )

    code_block_html = (
        f'<div class="pyreball-block-fit-content pyreball-centered">{sep}'
        f"{source_code_str}{sep}"
        f"</div>"
    )

    if caption_position == "top":
        code_block_html = caption_element + sep + code_block_html
    else:
        code_block_html = code_block_html + sep + caption_element

    code_block_html = (
        f'<div class="pyreball-block-fit-content {ALIGN_CLASS_MAP[align]}">{sep}'
        f"{code_block_html}{sep}</div>"
    )

    code_block_html = (
        f'<div class="pyreball-code-wrapper">{sep}{code_block_html}{sep}</div>'
    )

    return code_block_html


def print_code_block(
    *values: Any,
    caption: Optional[str] = None,
    reference: Optional[Reference] = None,
    align: Optional[str] = None,
    caption_position: Optional[str] = None,
    numbered: Optional[bool] = None,
    cl: ClParameter = None,
    attrs: AttrsParameter = None,
    pre_cl: ClParameter = None,
    pre_attrs: AttrsParameter = None,
    sep: str = "",
    end: str = "\n",
    syntax_highlight: Optional[Literal["python"]] = "python",
) -> None:
    """
    Print values as a source code into a preformatted block.

    This element is used to display a source code in a block.
    It is possible to highlight the code syntax by setting `syntax_highlight`
    parameter to an appropriate string.

    Args:
        *values: Zero or more values to be enclosed in the tag.
            All values are converted to strings.
        caption: Text caption.
        reference: Reference object.
        align: How to align the code block horizontally.
            Acceptable values are `'left'`, `'center'`, and `'right'`.
            Defaults to settings from config or CLI arguments if `None`.
        caption_position: Where to place the caption.
            Acceptable values are `'top'`, and `'bottom'`.
            Defaults to settings from config or CLI arguments if `None`.
        numbered: Whether the caption should be numbered.
            Defaults to settings from config or CLI arguments if `None`.
        cl: One or more class names to be added to the `<code>` tag.
            If a string is provided, it is used as it is.
            If a list of strings is provided, the strings are joined with space.
            If `None`, no class is added.
            If an empty list is provided, class attribute is added with an empty string.
        attrs: Additional attributes to be added to the `<code>` tag.
            Dictionary `{"key1": "value1", ..., "keyN": "valueN"}`
            is converted to `key1="value1" ... keyN="valueN"`.
            To construct boolean HTML attributes, set `None` for given key.
            Any quotes in values are not escaped.
        pre_cl: The same as `cl` parameter, but for the `<pre>` tag.
        pre_attrs: The same as `attrs` parameter, but for the `<pre>` tag.
        sep: String separator of the values inside the tag. Defaults to an empty string.
        end: String appended after the tag. Defaults to a newline.
        syntax_highlight: Syntax highlighting language.
            Currently only `'python'` is supported. If `None`,
            no highlight is applied. When highlight is turned on,
            language name is added to the `<code>` element as a class.
    """

    source_code_str = code_block(
        *values,
        cl=cl,
        attrs=attrs,
        pre_cl=pre_cl,
        pre_attrs=pre_attrs,
        sep=sep,
        syntax_highlight=syntax_highlight,
    )
    if not get_parameter_value("html_file_path") or get_parameter_value("keep_stdout"):
        builtins.print(source_code_str)
    if get_parameter_value("html_file_path"):
        if "code_block_index" not in _code_block_memory:
            _code_block_memory["code_block_index"] = 1
        code_block_index = _code_block_memory["code_block_index"]

        align = cast(
            str,
            merge_values(
                primary_value=align,
                secondary_value=get_parameter_value("align_code_blocks"),
            ),
        )
        caption_position = cast(
            str,
            merge_values(
                primary_value=caption_position,
                secondary_value=get_parameter_value("code_block_captions_position"),
            ),
        )
        numbered = bool(
            merge_values(
                primary_value=numbered,
                secondary_value=get_parameter_value("numbered_code_blocks"),
            )
        )

        code_block_html = _wrap_code_block_html(
            source_code_str=source_code_str,
            code_block_index=code_block_index,
            caption=caption,
            reference=reference,
            align=align,
            caption_position=caption_position,
            numbered=numbered,
            sep=sep,
        )
        _write_to_html(code_block_html, end=end)

        _code_block_memory["code_block_index"] += 1


def print(*values: Any, sep: str = "", end: str = "\n") -> None:
    """Print values as strings to HTML file.

    Args:
        *values: Zero or more values to be printed.
            Each value is converted to a string first.
        sep: Separator string to concatenate the values with.
            Defaults to an empty space.
        end: String appended after the values. Defaults to a newline.
    """
    str_values = map(str, values)
    string = sep.join(str_values) + end
    if not get_parameter_value("html_file_path") or get_parameter_value("keep_stdout"):
        builtins.print(string)
    if get_parameter_value("html_file_path"):
        _write_to_html(string, end="")


def _prepare_caption_element(
    prefix: str, caption: Optional[str], numbered: bool, index: int, anchor_link: str
) -> str:
    if numbered:
        caption_text = f"{prefix} {index}"
        if caption:
            caption_text = f"{caption_text}: {caption}"
        else:
            caption_text = f"{caption_text}."
    elif caption:
        caption_text = f"{caption}"
    else:
        caption_text = ""
    return (
        f'\n<div class="pyreball-text-centered">'
        f'<a id="{anchor_link}"><b>\n{caption_text}\n</b></a>'
        f"</div>\n"
    )


def _compute_length_menu_for_datatables(
    paging_sizes: List[Union[int, str]]
) -> Tuple[List[int], List[Union[int, str]]]:
    arr_1: List[int] = []
    arr_2: List[Union[int, str]] = []
    for size in paging_sizes:
        if isinstance(size, int):
            arr_1.append(size)
            arr_2.append(size)
        elif isinstance(size, str) and size.lower() == "all":
            arr_1.append(-1)
            arr_2.append(size)
        else:
            raise ValueError(f"Unsupported value in paging_sizes: {size}")
    return arr_1, arr_2


def _gather_datatables_setup(
    display_option: str = "full",
    scroll_y_height: str = "300px",
    scroll_x: bool = True,
    sortable: bool = False,
    sorting_definition: Optional[List[Tuple[int, str]]] = None,
    paging_sizes: Optional[List[Union[int, str]]] = None,
    search_box: bool = False,
    col_align_def: Optional[List[Dict[str, Any]]] = None,
    datatables_definition: Optional[Dict[str, Any]] = None,
) -> Optional[Dict[str, Any]]:
    if datatables_definition is not None:
        return datatables_definition

    datatables_setup: Dict[str, Any] = {}
    if display_option == "scrolling":
        datatables_setup["paging"] = False
        datatables_setup["scrollCollapse"] = True
        datatables_setup["scrollY"] = scroll_y_height
    elif display_option == "paging":
        datatables_setup["paging"] = True
        if paging_sizes is None:
            paging_sizes = [10, 25, 100, "All"]
        datatables_setup["lengthMenu"] = _compute_length_menu_for_datatables(
            paging_sizes
        )
    elif display_option == "full":
        datatables_setup["paging"] = False
    if scroll_x:
        datatables_setup["scrollX"] = True

    datatables_setup["info"] = False

    datatables_setup["searching"] = search_box

    if sortable or sorting_definition is not None:
        if sorting_definition is None:
            datatables_setup["order"] = []
        else:
            datatables_setup["order"] = sorting_definition
    else:
        datatables_setup["ordering"] = False

    if col_align_def is not None:
        datatables_setup["columnDefs"] = col_align_def

    return datatables_setup


def _check_col_alignment_value(value: str):
    allowed_values = ["left", "center", "right"]
    if value not in allowed_values:
        raise ValueError(
            "col_align must use only the following values: "
            f"{', '.join(allowed_values)}."
        )


def _prepare_col_alignment_definition(
    df: "pandas.DataFrame",
    col_align: Optional[Union[str, List[str]]] = None,
    index: bool = True,
) -> List[Dict[str, Any]]:
    if index:
        # when index is shown, we must also work with index columns
        df = df.reset_index()

    if col_align is None:
        # Pyreball's default alignment
        col_align = ["left"] * df.shape[1]
        for col in df.select_dtypes(include="number").columns:
            col_align[df.columns.get_loc(col)] = "right"
    elif isinstance(col_align, str):
        # All columns have the same alignment
        _check_col_alignment_value(col_align)
        return [{"targets": list(range(df.shape[1])), "className": f"dt-{col_align}"}]
    elif len(col_align) != df.shape[1]:
        # Alignment definition is a list and the length does not match
        raise ValueError(
            "col_align list must have the same length as the column list, "
            "including any index columns if they are shown too."
        )
    else:
        # it's a list with correct length, just check its values
        list(map(_check_col_alignment_value, col_align))

    # create alignment mapping in form <align_value> -> <col_indices>
    mapping: Dict[str, List[int]] = {}
    for i, al in enumerate(col_align):
        mapping.setdefault(al, []).append(i)

    # build final col definition for DataTables
    col_defs = []
    for key, value in mapping.items():
        col_defs.append({"targets": value, "className": f"dt-{key}"})
    return col_defs


def _prepare_table_html(
    df: "pandas.DataFrame",
    tab_index: int = 0,
    caption: Optional[str] = None,
    reference: Optional[Reference] = None,
    align: str = "center",
    caption_position: str = "top",
    numbered: bool = True,
    col_align: Optional[Union[str, List[str]]] = None,
    display_option: str = "full",
    scroll_y_height: str = "300px",
    scroll_x: bool = True,
    sortable: bool = False,
    sorting_definition: Optional[List[Tuple[int, str]]] = None,
    paging_sizes: Optional[List[Union[int, str]]] = None,
    search_box: bool = False,
    datatables_style: Union[str, List[str]] = "display",
    datatables_definition: Optional[Dict[str, Any]] = None,
    **kwargs: Any,
) -> str:
    table_classes = []
    if isinstance(datatables_style, list):
        table_classes += datatables_style
    else:
        table_classes.append(datatables_style)

    col_align_def = _prepare_col_alignment_definition(
        df=df, col_align=col_align, index=kwargs.get("index", True)
    )

    datatables_setup = _gather_datatables_setup(
        display_option=display_option,
        scroll_y_height=scroll_y_height,
        scroll_x=scroll_x,
        sortable=sortable,
        sorting_definition=sorting_definition,
        paging_sizes=paging_sizes,
        search_box=search_box,
        col_align_def=col_align_def,
        datatables_definition=datatables_definition,
    )

    if "border" not in kwargs:
        # If border is not set explicitly,
        # turn it off because it would clash with datatables styling
        kwargs["border"] = 0

    kwargs["sparsify"] = False
    df_html = df.to_html(classes=table_classes, **kwargs)
    if reference:
        _check_and_mark_reference(reference)
        anchor_link = f"table-{reference.id}-{tab_index}"
    else:
        anchor_link = f"table-{tab_index}"

    caption_element = _prepare_caption_element(
        prefix="Table",
        caption=caption,
        numbered=numbered,
        index=tab_index,
        anchor_link=anchor_link,
    )

    table_wrapper_inner_id = "pyreball-table-wrapper-inner-" + str(tab_index)
    table_html = (
        f'<div id="{table_wrapper_inner_id}" '
        f'class="pyreball-block-fit-content pyreball-centered">'
        f"{df_html}\n"
        f"</div>"
    )

    if caption_position == "top":
        table_html = caption_element + table_html
    else:
        table_html = table_html + caption_element

    table_html = (
        f'<div class="pyreball-block-fit-content {ALIGN_CLASS_MAP[align]}">'
        f"{table_html}</div>"
    )

    table_html = f'<div class="pyreball-table-wrapper">\n{table_html}\n</div>'

    if datatables_setup is not None:
        table_init = json.dumps(datatables_setup)
        js = f"new DataTable('#{table_wrapper_inner_id} > table', {table_init});"
        table_html += f"\n<script>{js}</script>"

    return table_html


def _parse_tables_paging_sizes(sizes: str) -> List[Union[int, str]]:
    return [
        value if value.lower() == "all" else int(value) for value in sizes.split(",")
    ]


def print_table(
    df: "pandas.DataFrame",
    caption: Optional[str] = None,
    reference: Optional[Reference] = None,
    align: Optional[str] = None,
    caption_position: Optional[str] = None,
    numbered: Optional[bool] = None,
    col_align: Optional[Union[str, List[str]]] = None,
    display_option: Optional[str] = None,
    paging_sizes: Optional[List[Union[int, str]]] = None,
    scroll_y_height: Optional[str] = None,
    scroll_x: Optional[bool] = None,
    sortable: Optional[bool] = None,
    sorting_definition: Optional[List[Tuple[int, str]]] = None,
    search_box: Optional[bool] = None,
    datatables_style: Optional[Union[str, List[str]]] = None,
    datatables_definition: Optional[Dict[str, Any]] = None,
    **kwargs: Any,
) -> None:
    """Print pandas DataFrame into HTML.

    It uses DataTables JavaScript library to display the table.

    Args:
        df: Data frame to be printed.
        caption: Text caption.
        reference: Reference object.
        align: How to align the table horizontally.
            Acceptable values are `'left'`, `'center'`, and `'right'`.
            Defaults to settings from config or CLI arguments if `None`.
        caption_position: Where to place the caption.
            Acceptable values are `'top'`, and `'bottom'`.
            Defaults to settings from config or CLI arguments if `None`.
        numbered: Whether the caption should be numbered.
            Defaults to settings from config or CLI arguments if `None`.
        col_align: Alignment of individual columns. Can be provided as a list
            of values `'left'`, `'center'`, and `'right'`. The list must match the
            number of columns, including the index when displayed.
            When provided as a simple string, this value is used for all columns.
            When `None`, default Pyreball alignment is used, i.e. numeric columns
            are right-aligned and all other columns left-aligned.
        display_option: How to display the table. This option is useful for long tables,
            which should not be displayed fully. Acceptable values are:
            `'full'` (show the full table), `'scrolling'` (show the table
            in scrolling mode on y-axis), `'paging'` (show the table in paging mode).
            Defaults to settings from config or CLI arguments if `None`.
        paging_sizes: A list of page sizes to display in paging mode.
            Allowed values in the list are integer values and string
            `'All'` (the case is not important).
            When `display_option` is not `'paging'`, the value is ignored.
            Defaults to settings from config or CLI arguments if `None`.
            If it still remains `None`, values `[10, 25, 100, "All"]` are used.
        scroll_y_height: Height of the tables when `display_option` is set to
            `'scrolling'`. Any string compatible with CSS sizing can be used,
            e.g. `'300px'`, `'20em'`, etc. Ignored with other display options.
            Defaults to settings from config or CLI arguments if `None`.
        scroll_x: Whether to allow scrolling on the x-axis. If set to `False`,
            a wide table is allowed to overflow the main container.
            It is recommended to set this to `True`, especially with
            `display_option='scrolling'`, because otherwise the table header
            won't interact properly when scrolling horizontally.
        sortable: Whether to allow sortable columns.
            Defaults to settings from config or CLI arguments if `None`.
        sorting_definition: How to sort the table columns initially,
            in the form of a list of tuples `(<column_index>, <sorting>)`,
            where `<sorting>` is either `'asc'` or `'desc'`.
            When `None`, the columns are not pre-sorted.
        search_box: Whether to show the search box for the table.
            Defaults to settings from config or CLI arguments if `None`.
        datatables_style: One or more class names for styling tables using
            Datatables styling. See https://datatables.net/manual/styling/classes
            for possible values. Can be either a string with the class name,
            or a list of class name strings.
        datatables_definition: Custom setup for datatables in the form of a dictionary.
            This dictionary is serialized to json and passed to `DataTable` JavaScript
            object as it is. If set (i.e. not None), values of parameters
            `col_align`, `display_option`, `paging_sizes`, `scroll_y_height`,
            `scroll_x`, `sortable`, `sorting_definition`, and `search_box` are ignored.
            Note that `datatables_style` is independent of this parameter.
        **kwargs: Other parameters to pandas `to_html()` method. Note that parameter
            `sparsify` is explicitly set to `False` by Pyreball, because tables
            with multi-index would not be displayed correctly using DataTables library.
    """
    if not get_parameter_value("html_file_path") or get_parameter_value("keep_stdout"):
        builtins.print(df)
    if get_parameter_value("html_file_path"):
        if "table_index" not in _table_memory:
            _table_memory["table_index"] = 1
        table_index = _table_memory["table_index"]

        align = cast(
            str,
            merge_values(
                primary_value=align, secondary_value=get_parameter_value("align_tables")
            ),
        )
        caption_position = cast(
            str,
            merge_values(
                primary_value=caption_position,
                secondary_value=get_parameter_value("table_captions_position"),
            ),
        )
        numbered = bool(
            merge_values(
                primary_value=numbered,
                secondary_value=get_parameter_value("numbered_tables"),
            )
        )
        display_option = str(
            merge_values(
                primary_value=display_option,
                secondary_value=get_parameter_value("tables_display_option"),
            )
        )
        scroll_y_height = str(
            merge_values(
                primary_value=scroll_y_height,
                secondary_value=get_parameter_value("tables_scroll_y_height"),
            )
        )
        scroll_x = bool(
            merge_values(
                primary_value=scroll_x,
                secondary_value=get_parameter_value("tables_scroll_x"),
            )
        )
        sortable = bool(
            merge_values(
                primary_value=sortable,
                secondary_value=get_parameter_value("sortable_tables"),
            )
        )
        paging_sizes = merge_values(
            primary_value=paging_sizes,
            secondary_value=_parse_tables_paging_sizes(
                get_parameter_value("tables_paging_sizes")
            ),
        )
        search_box = bool(
            merge_values(
                primary_value=search_box,
                secondary_value=get_parameter_value("tables_search_box"),
            )
        )
        conf_datatables_style = get_parameter_value("tables_datatables_style")
        if conf_datatables_style:
            conf_datatables_style = re.split(r"[, ]", conf_datatables_style)
        datatables_style = cast(
            Union[str, List[str]],
            merge_values(
                primary_value=datatables_style,
                secondary_value=conf_datatables_style,
            ),
        )

        table_html = _prepare_table_html(
            df=df,
            tab_index=table_index,
            caption=caption,
            reference=reference,
            align=align,
            caption_position=caption_position,
            numbered=numbered,
            col_align=col_align,
            display_option=display_option,
            scroll_y_height=scroll_y_height,
            scroll_x=scroll_x,
            sortable=sortable,
            sorting_definition=sorting_definition,
            paging_sizes=paging_sizes,
            search_box=search_box,
            datatables_style=datatables_style,
            datatables_definition=datatables_definition,
            **kwargs,
        )
        _write_to_html(table_html)
        _table_memory["table_index"] += 1


def _construct_image_anchor_link(reference: Optional[Reference], fig_index: int) -> str:
    if reference:
        _check_and_mark_reference(reference)
        return f"img-{reference.id}-{fig_index}"
    else:
        return f"img-{fig_index}"


def _wrap_image_element_by_outer_divs(
    img_element: str, align: str, img_type: str
) -> str:
    img_element = (
        f'<div align="{align}"><div style="display: inline-block;">'
        f"{img_element}"
        f"</div></div>"
    )
    wrapper_classes = f"pyreball-image-wrapper pyreball-{img_type}-fig"
    return f'<div class="{wrapper_classes}">{img_element}</div>'


def _prepare_matplotlib_image_element(
    fig: "matplotlib.figure.Figure",
    fig_index: int,
    image_format: Optional[str] = None,
    embedded: Optional[bool] = None,
) -> str:
    if image_format is not None and image_format not in ["svg", "png"]:
        raise ValueError('Matplotlib format can be only "svg" or "png".')

    image_format = merge_values(
        primary_value=image_format,
        secondary_value=get_parameter_value("matplotlib_format"),
    )
    embedded = merge_values(
        primary_value=embedded,
        secondary_value=get_parameter_value("matplotlib_embedded"),
    )

    if embedded:
        if image_format == "svg":
            f = io.BytesIO()
            fig.savefig(f, format="svg")
            img_element = f.getvalue().decode("utf-8")
        else:
            raise ValueError(
                "Only svg format can be used for embedded matplotlib figures."
            )
    elif get_parameter_value("html_dir_path") and get_parameter_value("html_dir_name"):
        make_sure_dir_exists(get_parameter_value("html_dir_path"))
        img_file_name = f"img_{fig_index:03d}.{image_format}"
        fig.savefig(
            os.path.join(get_parameter_value("html_dir_path"), img_file_name),
            format=image_format,
            bbox_inches="tight",
        )
        img_element = (
            f'<img src="'
            f'{os.path.join(get_parameter_value("html_dir_name"), img_file_name)}'
            f'">'
        )
    else:
        raise RuntimeError("Failed to create a matplotlib image.")
    return img_element


def _prepare_altair_image_element(fig: AltairFigType, fig_index: int) -> str:
    vis_id = "altairvis" + str(fig_index)
    img_element = (
        f'<div id="{vis_id}">'
        f'</div><script type="text/javascript">\nvar spec = {fig.to_json()};\n'
    )
    img_element += 'var opt = {"renderer": "canvas", "actions": false};\n'
    img_element += f'vegaEmbed("#{vis_id}", spec, opt);'
    img_element += "</script>"
    return img_element


def _prepare_plotly_image_element(fig: "plotly.graph_objs.Figure") -> str:
    return fig.to_html(full_html=False, include_plotlyjs=False)


def _prepare_bokeh_image_element(fig: "bokeh.plotting._figure.figure") -> str:
    # noinspection PyPackageRequirements
    from bokeh.embed import components  # type: ignore

    script, div = components(fig)
    return f"<div>{div}{script}</div>"


def _prepare_image_element(
    fig: FigType,
    fig_index: int,
    matplotlib_format: Optional[str] = None,
    embedded: Optional[bool] = None,
) -> Tuple[str, str]:
    # Create the html string according to the figure type.
    # (if we checked type of fig, we would have to add the libraries to requirements)
    if (
        type(fig).__name__ == "Figure" and type(fig).__module__ == "matplotlib.figure"
    ) or (
        type(fig).__name__ == "PairGrid" and type(fig).__module__ == "seaborn.axisgrid"
    ):
        img_element = _prepare_matplotlib_image_element(
            fig=fig,
            fig_index=fig_index,
            image_format=matplotlib_format,
            embedded=embedded,
        )
        img_type = "matplotlib"
    elif type(fig).__name__ in [
        "Chart",
        "ConcatChart",
        "FacetChart",
        "HConcatChart",
        "LayerChart",
        "RepeatChart",
        "VConcatChart",
    ]:
        img_element = _prepare_altair_image_element(fig=fig, fig_index=fig_index)
        img_type = "altair"
    elif (
        type(fig).__name__ == "Figure"
        and type(fig).__module__ == "plotly.graph_objs._figure"
    ):
        img_element = _prepare_plotly_image_element(fig=fig)
        img_type = "plotly"
    elif type(fig).__name__.lower() == "figure" and type(fig).__module__ in [
        "bokeh.plotting.figure",
        "bokeh.plotting._figure",
    ]:
        img_element = _prepare_bokeh_image_element(fig=fig)
        img_type = "bokeh"
    else:
        raise ValueError(f"Unknown figure type {type(fig)}.")

    return img_element, img_type


def _print_figure(
    fig: FigType,
    caption: Optional[str] = None,
    reference: Optional[Reference] = None,
    align: str = "center",
    caption_position: str = "bottom",
    numbered: bool = True,
    matplotlib_format: Optional[str] = None,
    embedded: Optional[bool] = None,
) -> None:
    if not get_parameter_value("html_file_path"):
        # only when we don't print to HTML
        if type(fig).__name__.lower() == "figure" and type(fig).__module__ in [
            "bokeh.plotting.figure",
            "bokeh.plotting._figure",
        ]:
            from bokeh.plotting import show

            show(fig)
        else:
            fig.show()
    else:
        if "fig_index" not in _graph_memory:
            _graph_memory["fig_index"] = 1
        fig_index = _graph_memory["fig_index"]

        anchor_link = _construct_image_anchor_link(
            reference=reference, fig_index=fig_index
        )
        img_element, img_type = _prepare_image_element(
            fig=fig,
            fig_index=fig_index,
            matplotlib_format=matplotlib_format,
            embedded=embedded,
        )
        caption_element = _prepare_caption_element(
            prefix="Figure",
            caption=caption,
            numbered=numbered,
            index=fig_index,
            anchor_link=anchor_link,
        )

        if caption_position == "bottom":
            img_with_caption = img_element + caption_element
        elif caption_position == "top":
            img_with_caption = caption_element + img_element
        else:
            raise ValueError(
                f"caption_position must be 'top' or 'bottom', not {caption_position}."
            )
        img_html = _wrap_image_element_by_outer_divs(
            img_element=img_with_caption,
            align=align,
            img_type=img_type,
        )

        _write_to_html(img_html)
        _graph_memory["fig_index"] += 1


def print_figure(
    fig: FigType,
    caption: Optional[str] = None,
    reference: Optional[Reference] = None,
    align: Optional[str] = None,
    caption_position: Optional[str] = None,
    numbered: Optional[bool] = None,
    matplotlib_format: Optional[str] = None,
    embedded: Optional[bool] = None,
) -> None:
    """
    Print a figure.

    Args:
        fig: A figure object.
        caption: Caption of the figure.
        reference: Reference object for link creation.
        align: How to align the graph horizontally.
            Acceptable values are `'left'`, `'center'`, and `'right'`.
            Defaults to settings from config or CLI arguments if `None`.
        caption_position: Where to place the caption.
            Acceptable values are `'top'`, and `'bottom'`.
            Defaults to settings from config or CLI arguments if `None`.
        numbered: Whether the caption should be numbered.
            Defaults to settings from config or CLI arguments if `None`.
        matplotlib_format: Format for matplotlib figures.
            Acceptable values are `'svg'`, and `'png'`.
            Defaults to settings from config or CLI arguments if `None`.
        embedded: Whether to embed the figure directly into HTML;
            Only applicable for matplotlib svg images.
            Defaults to settings from config or CLI arguments if `None`.
    """

    align = cast(
        str,
        merge_values(
            primary_value=align, secondary_value=get_parameter_value("align_figures")
        ),
    )
    caption_position = cast(
        str,
        merge_values(
            primary_value=caption_position,
            secondary_value=get_parameter_value("figure_captions_position"),
        ),
    )
    numbered = bool(
        merge_values(
            primary_value=numbered,
            secondary_value=get_parameter_value("numbered_figures"),
        )
    )

    _print_figure(
        fig=fig,
        caption=caption,
        reference=reference,
        align=align,
        caption_position=caption_position,
        numbered=numbered,
        matplotlib_format=matplotlib_format,
        embedded=embedded,
    )



import os
from pathlib import Path

import pkg_resources
import pytest
from pyreball.__main__ import (
    _contains_class,
    _get_config_directory,
    _get_output_dir_and_file_stem,
    _insert_heading_title_and_toc,
    _insert_js_and_css_links,
    _parse_heading_info,
    _replace_ids,
    parse_arguments,
)
from pyreball.constants import (
    CONFIG_INI_FILENAME,
    HTML_TEMPLATE_FILENAME,
    LINKS_INI_FILENAME,
    STYLES_TEMPLATE_FILENAME,
)


@pytest.mark.parametrize(
    "lines,expected_result",
    [
        (
            [
                "<html>",
                "</html>",
            ],
            [
                "<html>",
                "</html>",
            ],
        ),
        (
            [
                "<html>",
                'Reference to table <a href="#ref-id99">id99</a>',
                'Reference to image <a href="#ref-id17">id17</a>',
                '<a name="img-id17-5">Image caption</a><a name="table-id99-1">Table caption</a>',
                'Reference to table <a href="#ref-id123">id123</a> and image <a href="#ref-id42">id42</a>',
                '<a name="table-id123-2">Another table caption</a>',
                '<a name="img-id42-2">Another image caption</a>',
                'Reference again to table <a href="#ref-id123">id123</a> and image <a href="#ref-id42">id42</a>',
                "</html>",
            ],
            [
                "<html>",
                'Reference to table <a href="#table-1">1</a>',
                'Reference to image <a href="#img-5">5</a>',
                '<a name="img-5">Image caption</a><a name="table-1">Table caption</a>',
                'Reference to table <a href="#table-2">2</a> and image <a href="#img-2">2</a>',
                '<a name="table-2">Another table caption</a>',
                '<a name="img-2">Another image caption</a>',
                'Reference again to table <a href="#table-2">2</a> and image <a href="#img-2">2</a>',
                "</html>",
            ],
        ),
        (
            [
                "<html>",
                'Reference to code block <a href="#ref-id99">id99</a>',
                '<a name="code-block-id99-1">Code caption</a>',
                "</html>",
            ],
            [
                "<html>",
                'Reference to code block <a href="#code-block-1">1</a>',
                '<a name="code-block-1">Code caption</a>',
                "</html>",
            ],
        ),
        (
            [
                "<html>",
                'Reference to chapter <a href="#ref-id123">id123</a>',
                '<h1 id="ch_id123_heading_1_1">My Chapter'
                '<a class="pyreball-anchor-link" href="#ch_id123_heading_1_1">\u00B6</a></h1>',
                'Reference to chapter <a href="#ref-id123">id123</a> again',
                "</html>",
            ],
            [
                "<html>",
                'Reference to chapter <a href="#ch_heading_1_1">My Chapter</a>',
                '<h1 id="ch_heading_1_1">My Chapter<a class="pyreball-anchor-link" href="#ch_heading_1_1">\u00B6</a></h1>',
                'Reference to chapter <a href="#ch_heading_1_1">My Chapter</a> again',
                "</html>",
            ],
        ),
    ],
)
def test__replace_ids(lines, expected_result):
    result = _replace_ids(lines)
    assert result == expected_result


@pytest.mark.parametrize(
    "test_input,expected_result",
    [
        (
            '<h2 id="result_of_addition_2">Result of addition<a class="pyreball-anchor-link" href="#result_of_addition_2">¶</a></h2>',
            (2, "result_of_addition_2", "Result of addition"),
        ),
        (
            '<h3 id="some_id">Whatever text - is necessary - 999'
            '<a class="pyreball-anchor-link" href="#result_of_addition_2">¶</a></h3>',
            (3, "some_id", "Whatever text - is necessary - 999"),
        ),
        (
            '<h3 id="some_id">Whatever text - is necessary - 999'
            '<a class="pyreball-anchor-link" href="#result_of_addition_2">¶</a></h3>',
            (3, "some_id", "Whatever text - is necessary - 999"),
        ),
        (
            '<h3 id="some_id">Whatever text - also <code>code</code> and <b><em>bold emphasis</em></b> - 999'
            '<a class="pyreball-anchor-link" href="#result_of_addition_2">¶</a></h3>',
            (3, "some_id", "Whatever text - also code and bold emphasis - 999"),
        ),
        ("<div>paragraph</div>", None),
    ],
)
def test_parse_heading_info(test_input, expected_result):
    assert _parse_heading_info(test_input) == expected_result


@pytest.mark.parametrize(
    "title_set",
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize("include_toc", [True, False])
def test__insert_heading_title_and_toc__with_headings(include_toc, title_set):
    if title_set:
        title = '<title class="custom_pyreball_title">Custom Title</title>'
        expected_toc_heading = "Custom Title"
    else:
        title = "<title>Default Title</title>"
        expected_toc_heading = "Table of Contents"

    lines = [
        "<html>",
        "<head>",
        title,
        "<head/>",
        "<body>",
        '<div class="pyreball-main-container">',
        '<h1 id="1_heading_h1_1">1  heading h1<a class="pyreball-anchor-link" href="#1_heading_h1_1">¶</a></h1>',
        '<h2 id="1_1_heading_h2_2">1.1  heading h2<a class="pyreball-anchor-link" href="#1_1_heading_h2_2">¶</a></h2>',
        '<h2 id="1_2_heading_h2_3">1.2  heading h2<a class="pyreball-anchor-link" href="#1_2_heading_h2_3">¶</a></h2>',
        '<h3 id="1_2_1_heading_h3_4">1.2.1  heading h3<a class="pyreball-anchor-link" href="#1_2_1_heading_h3_4">¶</a></h3>',
        '<h3 id="1_2_2_heading_h3_5">1.2.2  heading h3<a class="pyreball-anchor-link" href="#1_2_2_heading_h3_5">¶</a></h3>',
        '<h1 id="2_heading_h1_6">2  heading h1<a class="pyreball-anchor-link" href="#2_heading_h1_6">¶</a></h1>',
        '<h1 id="3_heading_h1_7">3  heading h1<a class="pyreball-anchor-link" href="#3_heading_h1_7">¶</a></h1>',
        '<h2 id="3_1_heading_h2_8">3.1  heading h2<a class="pyreball-anchor-link" href="#3_1_heading_h2_8">¶</a></h2>',
        "</div>",
        "</body>",
        "</html>",
    ]

    result = _insert_heading_title_and_toc(lines=lines, include_toc=include_toc)

    expected_title_and_toc = []
    if title_set and not include_toc:
        expected_title_and_toc = [
            f'<h1 id="toc_generated_0">{expected_toc_heading}<a class="pyreball-anchor-link" href="#toc_generated_0">¶</a></h1>\n'
        ]
    elif include_toc:
        expected_title_and_toc = [
            f'<h1 id="toc_generated_0">{expected_toc_heading}<a class="pyreball-anchor-link" href="#toc_generated_0">¶</a></h1>\n',
            '<a href="#1_heading_h1_1">1  heading h1</a><br/>\n',
            '<ul style="list-style-type:none; margin:0px">\n',
            '<li><a href="#1_1_heading_h2_2">1.1  heading h2</a></li>\n',
            '<li><a href="#1_2_heading_h2_3">1.2  heading h2</a></li>\n',
            '<ul style="list-style-type:none; margin:0px">\n',
            '<li><a href="#1_2_1_heading_h3_4">1.2.1  heading h3</a></li>\n',
            '<li><a href="#1_2_2_heading_h3_5">1.2.2  heading h3</a></li>\n',
            "</ul>\n",
            "</ul>\n",
            '<a href="#2_heading_h1_6">2  heading h1</a><br/>\n',
            '<a href="#3_heading_h1_7">3  heading h1</a><br/>\n',
            '<ul style="list-style-type:none; margin:0px">\n',
            '<li><a href="#3_1_heading_h2_8">3.1  heading h2</a></li>\n',
            "</ul>\n",
        ]

    expected_result = (
        [
            "<html>",
            "<head>",
            title,
            "<head/>",
            "<body>",
            '<div class="pyreball-main-container">',
        ]
        + expected_title_and_toc
        + [
            '<h1 id="1_heading_h1_1">1  heading h1<a class="pyreball-anchor-link" href="#1_heading_h1_1">¶</a></h1>',
            '<h2 id="1_1_heading_h2_2">1.1  heading h2<a class="pyreball-anchor-link" href="#1_1_heading_h2_2">¶</a></h2>',
            '<h2 id="1_2_heading_h2_3">1.2  heading h2<a class="pyreball-anchor-link" href="#1_2_heading_h2_3">¶</a></h2>',
            '<h3 id="1_2_1_heading_h3_4">1.2.1  heading h3<a class="pyreball-anchor-link" href="#1_2_1_heading_h3_4">¶</a></h3>',
            '<h3 id="1_2_2_heading_h3_5">1.2.2  heading h3<a class="pyreball-anchor-link" href="#1_2_2_heading_h3_5">¶</a></h3>',
            '<h1 id="2_heading_h1_6">2  heading h1<a class="pyreball-anchor-link" href="#2_heading_h1_6">¶</a></h1>',
            '<h1 id="3_heading_h1_7">3  heading h1<a class="pyreball-anchor-link" href="#3_heading_h1_7">¶</a></h1>',
            '<h2 id="3_1_heading_h2_8">3.1  heading h2<a class="pyreball-anchor-link" href="#3_1_heading_h2_8">¶</a></h2>',
            "</div>",
            "</body>",
            "</html>",
        ]
    )

    assert result == expected_result


@pytest.mark.parametrize(
    "title_set",
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize("include_toc", [True, False])
def test__insert_heading_title_and_toc__without_headings(include_toc, title_set):
    if title_set:
        title = '<title class="custom_pyreball_title">Custom Title</title>'
    else:
        title = "<title>Default Title</title>"

    lines = [
        "<html>",
        "<head>",
        title,
        "<head/>",
        "<body>",
        '<div class="pyreball-main-container">',
        "</div>",
        "</body>",
        "</html>",
    ]

    result = _insert_heading_title_and_toc(lines=lines, include_toc=include_toc)

    expected_title_and_toc = []
    if title_set:
        expected_title_and_toc = [
            f'<h1 id="toc_generated_0">Custom Title<a class="pyreball-anchor-link" href="#toc_generated_0">¶</a></h1>\n'
        ]

    report_after = (
        [
            "<html>",
            "<head>",
            title,
            "<head/>",
            "<body>",
            '<div class="pyreball-main-container">',
        ]
        + expected_title_and_toc
        + [
            "</div>",
            "</body>",
            "</html>",
        ]
    )

    assert result == report_after


@pytest.mark.parametrize(
    "html_text,class_name,expected_result",
    [
        ("", "inline", False),
        ('<div class="inline"></div>', "inline", True),
        ('<div class="another"></div>', "another", True),
        ("<div class='inline'></div>", "inline", True),
        ('<div class="  hi inline ok  "></div>', "inline", True),
        ('<div class = "  hi \n inline ok \n  "></div>', "inline", True),
        ('<div class="  hi ok  ">inline</div>', "inline", False),
        ('<div class="  hi ok  " inline></div>', "inline", False),
        ('<div class="" inline></div>', "inline", False),
    ],
)
def test__contains_class(html_text, class_name, expected_result):
    assert _contains_class(html_text, class_name) == expected_result


@pytest.mark.parametrize(
    "html_content,external_links,expected_result",
    [
        (
            "",
            {"bokeh": ["l1", "l2"], "altair": ["l3"]},
            "",
        ),
        (
            (
                "<html><!--PYREBALL_HEAD_LINKS-->"
                '<div class="pyreball-bokeh-fig">'
                "</div></html>"
            ),
            {"bokeh": ["l1", "l2"], "altair": ["l3"]},
            '<html>l1\nl2<div class="pyreball-bokeh-fig"></div></html>',
        ),
        (
            (
                "<html><!--PYREBALL_HEAD_LINKS-->"
                '<div class="pyreball-altair-fig">'
                "</div>"
                '<div class="pyreball-table-wrapper">'
                "</div>"
                '<div class="pyreball-plotly-fig">'
                "</div>"
                '<div class="pyreball-code-wrapper">'
                "</div>"
                "</html>"
            ),
            {
                "altair": ["l1", "l2"],
                "jquery": ["l4"],
                "highlight_js": ["l3"],
                "datatables": ["l5"],
                "plotly": ["l6"],
            },
            (
                "<html>l4\nl1\nl2\nl5\nl3\nl6"
                '<div class="pyreball-altair-fig">'
                "</div>"
                '<div class="pyreball-table-wrapper">'
                "</div>"
                '<div class="pyreball-plotly-fig">'
                "</div>"
                '<div class="pyreball-code-wrapper">'
                "</div>"
                "</html>"
            ),
        ),
    ],
)
def test__insert_js_and_css_links(html_content, external_links, expected_result):
    assert _insert_js_and_css_links(html_content, external_links) == expected_result


def test__get_config_directory__custom_path_does_not_exist(tmpdir):
    tmpdir = Path(tmpdir)
    config_dir = "my_config_dir"
    with pytest.raises(NotADirectoryError):
        _get_config_directory(config_dir_path=tmpdir / config_dir)


def test__get_config_directory__custom_path_with_incomplete_files(tmpdir):
    tmpdir = Path(tmpdir)
    config_dir = "my_config_dir"
    (tmpdir / config_dir).mkdir()
    (tmpdir / config_dir / CONFIG_INI_FILENAME).touch()
    with pytest.raises(FileNotFoundError):
        _get_config_directory(config_dir_path=tmpdir / config_dir)


def test__get_config_directory__home_path_with_incomplete_files(tmpdir, mocker):
    tmpdir = Path(tmpdir)
    fake_home_path = tmpdir / "my_home"
    mocker.patch.object(Path, "home", return_value=fake_home_path)
    (fake_home_path / ".pyreball").mkdir(parents=True)
    (fake_home_path / ".pyreball" / HTML_TEMPLATE_FILENAME).touch()
    with pytest.raises(FileNotFoundError):
        _get_config_directory(config_dir_path=None)


@pytest.mark.parametrize(
    "config_path,home_config_exists,expected_result_unresolved",
    [
        # Path from CLI argument is used
        (Path("my_config_dir/"), False, Path("my_config_dir")),
        # Path from CLI argument is used
        (Path("my_config_dir"), True, Path("my_config_dir")),
        # Path from home
        (None, True, Path("my_home/.pyreball")),
        # Path from installation directory
        (None, False, Path(pkg_resources.resource_filename("pyreball", "cfg"))),
    ],
)
def test__get_config_directory__valid_conditions(
    config_path, home_config_exists, expected_result_unresolved, tmpdir, mocker
):
    tmpdir = Path(tmpdir)
    fake_home_path = tmpdir / "my_home"
    mocker.patch.object(Path, "home", return_value=fake_home_path)

    # Setup
    os.chdir(tmpdir)
    required_filename = [
        CONFIG_INI_FILENAME,
        LINKS_INI_FILENAME,
        STYLES_TEMPLATE_FILENAME,
        HTML_TEMPLATE_FILENAME,
    ]
    config_dir = "my_config_dir"
    (tmpdir / config_dir).mkdir()
    for filename in required_filename:
        (tmpdir / config_dir / filename).touch()
    if home_config_exists:
        (fake_home_path / ".pyreball").mkdir(parents=True)
        for filename in required_filename:
            (fake_home_path / ".pyreball" / filename).touch()

    expected_result = expected_result_unresolved.resolve()

    # Test
    result = _get_config_directory(config_dir_path=config_path)
    assert result == expected_result


def test__get_output_dir_and_file_stem__input_file_does_not_exist(tmpdir):
    input_path = Path(tmpdir) / "input.py"
    with pytest.raises(FileNotFoundError):
        _get_output_dir_and_file_stem(input_path=input_path, output_path_str=None)


@pytest.mark.parametrize(
    "input_path," "output_path_str," "expected_output_dir," "expected_filename_stem",
    [
        (Path("script.py"), None, "", "script"),
        (Path("a/b/script.py"), None, "a/b", "script"),
        (Path("script.py"), Path("output.html"), "", "output"),
        (Path("script.py"), Path("x/y/output.html"), "x/y", "output"),
        (Path("a/b/script.py"), Path("x/y/output.html"), "x/y", "output"),
        (Path("script.py"), Path("x/y"), "x/y", "script"),
        (Path("script.py"), Path("x/y/"), "x/y", "script"),
        (Path("script.py"), Path("output.unsupported"), "output.unsupported", "script"),
    ],
)
def test__get_output_dir_and_file_stem__valid_inputs(
    input_path,
    output_path_str,
    expected_output_dir,
    expected_filename_stem,
    tmpdir,
):
    os.chdir(tmpdir)
    input_path.parents[0].mkdir(parents=True, exist_ok=True)
    input_path.touch()

    result_output_dir_path, result_filename_stem = _get_output_dir_and_file_stem(
        input_path, output_path_str
    )
    expected_output_dir_path = tmpdir / expected_output_dir
    assert result_output_dir_path == expected_output_dir_path
    assert result_filename_stem == expected_filename_stem


@pytest.mark.parametrize(
    "args",
    [
        # missing required input_path
        [],
        # Empty string as input_path
        [""],
        # wrong param value
        ["--numbered-figures", "hello", "scripts/report.py"],
        # empty string for path option
        # white-space string for path option
        ["--output-path", "\n    \n", "scripts/report.py"],
        # The same for --config-path
        ["--config-path", "\n    \n", "scripts/report.py"],
    ],
)
def test_parse_arguments__invalid_arguments(args):
    with pytest.raises(SystemExit):
        assert parse_arguments(args)


@pytest.mark.parametrize(
    "args,expected_non_empty_result",
    [
        # only the required argument
        (
            ["scripts/report.py"],
            {"input_path": Path("scripts/report.py")},
        ),
        # with some table parameters
        (
            [
                "--tables-display-option",
                "paging",
                "--tables-scroll-y-height",
                "500px",
                "--tables-paging-sizes",
                "20,all",
                "scripts/report.py",
            ],
            {
                "tables_display_option": "paging",
                "tables_scroll_y_height": "500px",
                "tables_paging_sizes": "20,all",
                "input_path": Path("scripts/report.py"),
            },
        ),
        # with path arguments
        (
            [
                "--output-path",
                "my_output.html",
                "--config-path",
                "dir/my_config/",
                "scripts/report.py",
            ],
            {
                "output_path": Path("my_output.html"),
                "config_path": Path("dir/my_config/"),
                "input_path": Path("scripts/report.py"),
            },
        ),
        # --output-path can also link to a directory
        (
            [
                "--output-path",
                "my_dir/my_subdir",
                "scripts/report.py",
            ],
            {
                "output_path": Path("my_dir/my_subdir"),
                "input_path": Path("scripts/report.py"),
            },
        ),
        (
            ["--align-tables", "left", "scripts/report.py"],
            {"align_tables": "left", "input_path": Path("scripts/report.py")},
        ),
        # now --align-tables is considered scripts-args
        (
            ["scripts/report.py", "--align-tables", "left"],
            {
                "input_path": Path("scripts/report.py"),
                "script_args": ["--align-tables", "left"],
            },
        ),
        # with script_args
        (
            ["scripts/report.py", "-p", "20", "img.png"],
            {
                "input_path": Path("scripts/report.py"),
                "script_args": ["-p", "20", "img.png"],
            },
        ),
        # wrong page-width will be parsed as it is, but fixed later
        (
            ["--page-width", "20", "scripts/report.py"],
            {"page_width": 20, "input_path": Path("scripts/report.py")},
        ),
    ],
)
def test_parse_arguments__valid_arguments(args, expected_non_empty_result):
    expected_result = {
        "toc": None,
        "align_code_blocks": None,
        "code_block_captions_position": None,
        "numbered_code_blocks": None,
        "align_tables": None,
        "table_captions_position": None,
        "numbered_tables": None,
        "tables_display_option": None,
        "tables_scroll_y_height": None,
        "tables_scroll_x": None,
        "sortable_tables": None,
        "tables_paging_sizes": None,
        "tables_search_box": None,
        "tables_datatables_style": None,
        "align_figures": None,
        "figure_captions_position": None,
        "numbered_figures": None,
        "matplotlib_format": None,
        "matplotlib_embedded": None,
        "numbered_headings": None,
        "page_width": None,
        "keep_stdout": None,
        "output_path": None,
        "config_path": None,
        "input_path": None,
        "script_args": [],
    }
    expected_result.update(**expected_non_empty_result)
    assert parse_arguments(args) == expected_result



import argparse
import configparser
import json
import os
import re
import shutil
import sys
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Callable, cast, Dict, List, Optional, Tuple, Union

from pyreball.utils.logger import get_logger

ParametersType = Dict[str, Optional[Union[str, int]]]

logger = get_logger()

_parameter_cache = {}


class Parameter(ABC):
    def __init__(self, option_string: str, help: str):
        self.option_string = option_string
        self._config_param_key = option_string.replace("--", "")
        self._param_key = self._config_param_key.replace("-", "_")
        self.help = help

    @property
    @abstractmethod
    def param_key(self):
        ...

    @property
    @abstractmethod
    def config_param_key(self):
        ...

    @abstractmethod
    def add_argument_to_parser(self, parser: argparse.ArgumentParser) -> None:
        ...

    @abstractmethod
    def check_and_fix_value(
        self,
        value: Any,
        none_allowed: bool,
        warning_messages: List[str],
        error_messages: List[str],
    ) -> Any:
        ...


class ChoiceParameter(Parameter):
    def __init__(self, option_string: str, choices: List[str], default: str, help: str):
        super().__init__(option_string=option_string, help=help)
        self.default = default
        self.choices = choices

    @property
    def param_key(self):
        return self._param_key

    @property
    def config_param_key(self):
        return self._config_param_key

    def add_argument_to_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(self.option_string, choices=self.choices, help=self.help)

    def check_and_fix_value(
        self,
        value: Any,
        none_allowed: bool,
        warning_messages: List[str],
        error_messages: List[str],
    ) -> Any:
        return check_choice_string_parameter(
            key=self.param_key,
            value=cast(Optional[str], value),
            value_choices=self.choices,
            default_value=self.default,
            none_allowed=none_allowed,
            warning_messages=warning_messages,
            error_messages=error_messages,
        )


class IntegerParameter(Parameter):
    def __init__(
        self,
        option_string: str,
        boundaries: Tuple[Optional[int], Optional[int]],
        default: int,
        help: str,
    ):
        super().__init__(option_string=option_string, help=help)
        self.boundaries = boundaries
        self.default = default

    @property
    def param_key(self):
        return self._param_key

    @property
    def config_param_key(self):
        return self._config_param_key

    def add_argument_to_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(self.option_string, type=int, help=self.help)

    def check_and_fix_value(
        self,
        value: Any,
        none_allowed: bool,
        warning_messages: List[str],
        error_messages: List[str],
    ) -> Any:
        return check_integer_within_range(
            key=self.param_key,
            value=value,
            low=self.boundaries[0],
            high=self.boundaries[1],
            none_allowed=none_allowed,
            warning_messages=warning_messages,
            error_messages=error_messages,
        )


class StringParameter(Parameter):
    def __init__(
        self,
        option_string: str,
        default: str,
        help: str,
        validation_function: Optional[
            Callable[
                [str, Optional[str], str, bool, List[str], List[str]], Optional[str]
            ]
        ] = None,
    ):
        super().__init__(option_string=option_string, help=help)
        self.default = default
        self.validation_function = validation_function

    @property
    def param_key(self):
        return self._param_key

    @property
    def config_param_key(self):
        return self._config_param_key

    def add_argument_to_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(self.option_string, type=str, help=self.help)

    def check_and_fix_value(
        self,
        value: Optional[str],
        none_allowed: bool,
        warning_messages: List[str],
        error_messages: List[str],
    ) -> Any:
        if self.validation_function is not None:
            return self.validation_function(
                self.param_key,
                value,
                self.default,
                none_allowed,
                warning_messages,
                error_messages,
            )
        else:
            return value


def _matches_paging_sizes_string(value: str) -> bool:
    return bool(re.match(r"^(all|[\d]+)(,(all|[\d]+))*$", value, re.IGNORECASE))


def check_paging_sizes_string_parameter(
    key: str,
    value: Optional[str],
    default_value: str,
    none_allowed: bool,
    warning_messages: List[str],
    error_messages: List[str],
) -> Optional[str]:
    if value is None and not none_allowed:
        warning_messages.append(
            f'Parameter {key} was not set, setting its value to "{default_value}".'
        )
        return default_value
    elif value is not None and not _matches_paging_sizes_string(value):
        error_messages.append(
            f"Parameter {key} is set to an unsupported value {value}. "
            "Allowed values are integers and string "
            "'all' (no matter the case of letters), "
            "written as a non-empty comma-separated list."
        )
    return value


def check_choice_string_parameter(
    key: str,
    value: Optional[str],
    value_choices: List[str],
    default_value: str,
    none_allowed: bool,
    warning_messages: List[str],
    error_messages: List[str],
) -> Optional[str]:
    if value is None and not none_allowed:
        warning_messages.append(
            f'Parameter {key} was not set, setting its value to "{default_value}".'
        )
        return default_value
    elif value is not None and value not in value_choices:
        error_messages.append(
            f"Parameter {key} is set to an unsupported value {value}, "
            f'only these values are allowed: {", ".join(value_choices)}.'
        )
    return value


def check_integer_within_range(
    key: str,
    value: Any,
    low: Optional[int],
    high: Optional[int],
    none_allowed: bool,
    warning_messages: List[str],
    error_messages: List[str],
) -> Any:
    if value is None and none_allowed:
        return value
    try:
        value_int = int(value)
    except (ValueError, TypeError):
        error_messages.append(f"Could not parse {key} parameter as an integer.")
        return value
    else:
        if low is not None and value_int < low:
            value_int = low
            warning_messages.append(
                f"Parameter {key} is less than {low}, setting it to {low}."
            )
        if high is not None and value_int > high:
            value_int = high
            warning_messages.append(
                f"Parameter {key} is more than {high}, setting it to {high}."
            )
    return value_int


def check_and_fix_parameters(
    parameters: ParametersType,
    parameter_specifications: List[Parameter],
    none_allowed: bool,
) -> ParametersType:
    # collect the config values
    new_parameters: ParametersType = {}
    warning_messages: List[str] = []
    error_messages: List[str] = []

    for param_spec in parameter_specifications:
        new_parameters[param_spec.param_key] = param_spec.check_and_fix_value(
            parameters.get(param_spec.param_key),
            none_allowed=none_allowed,
            warning_messages=warning_messages,
            error_messages=error_messages,
        )

    for msg in warning_messages:
        logger.warning(msg)
    for msg in error_messages:
        logger.error(msg)

    if len(error_messages) > 0:
        sys.exit(1)

    return new_parameters


def read_file_config(filename: str, directory: Path) -> configparser.ConfigParser:
    config_path = directory / filename
    if config_path.is_file():
        config = configparser.ConfigParser()
        config.read(config_path)
        return config
    else:
        logger.error(
            f"Could not find file {filename} in {directory}. "
            f"Try re-generating configs by "
            f"pyreball-generate-config command or re-installing pyreball."
        )
        sys.exit(1)


def get_file_config(
    filename: str, directory: Path, parameter_specifications: List[Parameter]
) -> ParametersType:
    config = read_file_config(filename, directory)
    section_name = "Parameters"
    if section_name not in config:
        logger.error(
            f"{section_name} section not found in {directory / filename} "
            f"configuration file. Fix the file or try re-installing pyreball."
        )
        sys.exit(1)

    # collect the config values and use underscore instead of dashes
    config_params: ParametersType = {
        param_spec.param_key: config[section_name].get(param_spec.config_param_key)
        for param_spec in parameter_specifications
    }
    return check_and_fix_parameters(
        parameters=config_params,
        parameter_specifications=parameter_specifications,
        none_allowed=False,
    )


def _parse_links_from_config(links_str: str) -> List[str]:
    return [x for x in links_str.split("\n") if x]


def get_external_links_from_config(
    filename: str, directory: Path
) -> Dict[str, List[str]]:
    config = read_file_config(filename, directory)
    section_name = "Links"
    if section_name not in config:
        logger.error(
            f"{section_name} section not found in {directory / filename} "
            f"configuration file. Fix the file or try re-installing pyreball."
        )
        sys.exit(1)

    links = {
        key: _parse_links_from_config(value)
        for key, value in config[section_name].items()
    }
    required_keys = {
        "altair",
        "bokeh",
        "datatables",
        "highlight_js",
        "jquery",
        "plotly",
    }
    if links.keys() != required_keys:
        logger.error(
            "Configuration with items must contain links for exactly these keys: "
            f"{', '.join(sorted(list(required_keys)))}."
        )
        sys.exit(1)
    return links


def merge_values(primary_value: Any, secondary_value: Any) -> Any:
    return primary_value if primary_value is not None else secondary_value


def merge_parameter_dictionaries(
    primary_parameters: ParametersType,
    secondary_parameters: ParametersType,
    parameter_specifications: List[Parameter],
) -> ParametersType:
    return {
        param_spec.param_key: merge_values(
            primary_value=primary_parameters.get(param_spec.param_key),
            secondary_value=secondary_parameters.get(param_spec.param_key),
        )
        for param_spec in parameter_specifications
    }


def _map_env_value(value):
    if value == "None":
        return None
    elif value == "yes":
        return True
    elif value == "no":
        return False
    else:
        return value


def get_parameter_value(key: str) -> Any:
    if "params" not in _parameter_cache:
        _parameter_cache["params"] = {
            k: _map_env_value(v)
            for k, v in json.loads(
                os.environ.get("_TMP_PYREBALL_GENERATOR_PARAMETERS", "{}")
            ).items()
        }
        if "html_dir_path" in _parameter_cache["params"]:
            _parameter_cache["params"]["html_dir_name"] = os.path.basename(
                _parameter_cache["params"]["html_dir_path"]
            )
            _parameter_cache["params"]["html_file_path"] = (
                _parameter_cache["params"]["html_dir_path"] + ".html"
            )
    return _parameter_cache["params"].get(key)


def make_sure_dir_exists(directory: Optional[str]) -> None:
    if directory and not os.path.exists(directory):
        os.makedirs(directory)


def carefully_remove_directory_if_exists(directory: Path) -> None:
    """Remove directory if it exists and contains only files of given type.

    :param directory: path to a directory.
    """
    if directory.exists():
        # check that the folder contains only png and jpg files.
        # Otherwise, raise an error.
        for filename in directory.iterdir():
            if not re.search(r".(png|jpg|svg)$", filename.name):
                raise ValueError(
                    f"Cannot delete the original html directory {directory}. "
                    f"It contains file {filename} and only png, jpg and svg "
                    f"files are allowed."
                )
        try:
            shutil.rmtree(directory)
        except OSError:
            # print also the directory so that we have enough info
            logger.error(f"Error raised when deleting directory: {directory}")
            raise


class Substitutor:
    """Multiple-string substitution class.

    It replaces strings for a different strings in a sequential order.
    It is possible to use regex as well.
    """

    def __init__(self, replacements: List[Tuple[str, str]]) -> None:
        self.patterns = [re.compile(r[0]) for r in replacements]
        self.replacements = [r[1] for r in replacements]

    def sub(self, text: str) -> str:
        for i in range(len(self.patterns)):
            text = self.patterns[i].sub(self.replacements[i], text)
        return text



import argparse
import json
import os
import re
import sys
import textwrap
import xml
from pathlib import Path
from typing import cast, Dict, List, Optional, Tuple, Union
from xml.dom.minidom import parseString

import pkg_resources

from pyreball.constants import (
    CONFIG_INI_FILENAME,
    DEFAULT_PATH_TO_CONFIG,
    HTML_TEMPLATE_FILENAME,
    LINKS_INI_FILENAME,
    STYLES_TEMPLATE_FILENAME,
)
from pyreball.utils.logger import get_logger
from pyreball.utils.param import (
    carefully_remove_directory_if_exists,
    check_and_fix_parameters,
    check_paging_sizes_string_parameter,
    ChoiceParameter,
    get_external_links_from_config,
    get_file_config,
    IntegerParameter,
    merge_parameter_dictionaries,
    StringParameter,
    Substitutor,
)
from pyreball.utils.template import get_css, get_html

logger = get_logger()


def _replace_ids(lines: List[str]) -> List[str]:
    """
    Replace IDs of HTML elements to create working anchors based on references.

    Args:
        lines: Lines of the html file.

    Returns:
        Updated lines of the html file.
    """
    # collect all ids in form of "table-N-M", "img-N-M"
    all_table_and_img_ids = set()
    chapter_text_replacemenets = []
    for line in lines:
        # we need to replace "table", "img", "code-block" and chapter IDs.
        results = re.findall(r"table-id[\d]+-[\d]+", line)
        if results:
            all_table_and_img_ids.update(results)
        results = re.findall(r"img-id[\d]+-[\d]+", line)
        if results:
            all_table_and_img_ids.update(results)
        results = re.findall(r"code-block-id[\d]+-[\d]+", line)
        if results:
            all_table_and_img_ids.update(results)
        # now collect heading references:
        results = re.findall(r"ch_id[\d]+_[^\"]+", line)
        if results:
            all_table_and_img_ids.update(results)
            # obtain also the heading text
            search_result_text = re.search(results[0] + r"\">([^<]+)<", line)
            link_text = search_result_text.group(1) if search_result_text else ""
            search_result_id = re.search(r"_(id[\d]+)_", results[0])
            link_id = search_result_id.group(1) if search_result_id else ""
            if link_id and link_text:
                chapter_text_replacemenets.append((f">{link_id}<", f">{link_text}<"))
    # Prepare all replacement definitions for a substitutor below
    replacements = []
    for element_id in all_table_and_img_ids:
        # Tables, images and code blocks:
        re_results = re.search(r"(.+)-(id\d+)-(\d+)", element_id)
        if re_results:
            # this must be first
            replacements.append(
                (
                    f"ref-{re_results.group(2)}",
                    f"{re_results.group(1)}-{re_results.group(3)}",
                )
            )
            # this must be second (because it would catch the first case as well)
            replacements.append(
                (
                    f"{re_results.group(2)}(-{re_results.group(3)})?",
                    re_results.group(3),
                )
            )

        # Headings
        re_results = re.search(r"ch_(id\d+)_(.+)", element_id)
        if re_results:
            # this must be first
            replacements.append(
                (
                    f"ref-{re_results.group(1)}",
                    f"ch_{re_results.group(2)}",
                )
            )
            # this must be second (because it would catch the first case as well)
            replacements.append((element_id, f"ch_{re_results.group(2)}"))
    # add also replacements for links to chapters
    replacements += chapter_text_replacemenets

    # replace all table-N-M with table-M and Table N with Table M
    substitutor = Substitutor(replacements=replacements)
    return [substitutor.sub(line) for line in lines]


def _get_node_text(node: xml.dom.minidom.Element) -> str:
    result = []
    for child in node.childNodes:
        if child.nodeType in (xml.dom.Node.TEXT_NODE, xml.dom.Node.CDATA_SECTION_NODE):
            result.append(child.data)
        else:
            result.extend(_get_node_text(child))
    return "".join(result)


def _parse_heading_info(line: str) -> Optional[Tuple[int, str, str]]:
    heading_pattern = r"<h(\d).+</h(\d)>"
    m = re.search(heading_pattern, line)
    if m:
        heading_level = m.group(1)

        doc = parseString(m.group(0))
        heading = doc.getElementsByTagName(f"h{heading_level}")[0]
        heading_id = heading.getAttribute("id")
        content = _get_node_text(heading).replace("¶", "")
        return int(heading_level), heading_id, content
    else:
        return None


def _insert_heading_title_and_toc(
    lines: List[str], include_toc: bool = True
) -> List[str]:
    # try to extract the title from <title> element:
    report_title = None
    for line in lines:
        m = re.match(r'^<title class="custom_pyreball_title">([^<]*)</title>$', line)
        if m:
            report_title = m.group(1)
            break

    # get all headings in the report
    container_start_index = 0
    headings = []
    for i, line in enumerate(lines):
        if '<div class="pyreball-main-container">' in line:
            container_start_index = i

        if include_toc:
            heading_info = _parse_heading_info(line)
            if heading_info:
                headings.append(heading_info)

    if len(headings) > 0 and report_title is None:
        # only when headings were collected (only when include_toc=True)
        # and there was not title set manually
        report_title = "Table of Contents"

    lines_index = container_start_index + 1

    # prepare new HTML lines with TOC
    if report_title is not None:
        lines.insert(
            lines_index,
            (
                f'<h1 id="toc_generated_0">{report_title}'
                f'<a class="pyreball-anchor-link" href="#toc_generated_0">¶</a></h1>\n'
            ),
        )
        lines_index += 1
    current_level = 1
    for h in headings:
        # do we need to add also <ul> ?
        while h[0] > current_level:
            lines.insert(lines_index, '<ul style="list-style-type:none; margin:0px">\n')
            lines_index += 1
            current_level += 1

        # do we need to add also </ul> ?
        while h[0] < current_level:
            lines.insert(lines_index, "</ul>\n")
            lines_index += 1
            current_level -= 1

        # prepare the line:
        if h[0] == 1:
            current_line = f'<a href="#{h[1]}">{h[2]}</a><br/>\n'
        else:
            current_line = f'<li><a href="#{h[1]}">{h[2]}</a></li>\n'
        lines.insert(lines_index, current_line)
        lines_index += 1

    # at the end, get back to level 1 if necessary
    while 1 < current_level:
        lines.insert(lines_index, "</ul>\n")
        lines_index += 1
        current_level -= 1

    return lines


def _contains_class(html_text: str, class_name: str) -> bool:
    """
    Check whether the given HTML text contains the given class name in any element.

    Args:
        html_text: HTML text.
        class_name: Class to be found.

    Returns:
        True if the HTML text contains the given class name.
    """
    pattern = (
        r'class\s*=\s*["\']\s*(?:\S+\s+)*'
        + re.escape(class_name)
        + r'(?:\s+\S+)*\s*["\']'
    )
    return re.search(pattern, html_text) is not None


def _insert_js_and_css_links(
    html_content: str, external_links: Dict[str, List[str]]
) -> str:
    groups_of_links_to_add = set()
    add_jquery = False
    if _contains_class(
        html_text=html_content, class_name="inline-highlight"
    ) or _contains_class(html_text=html_content, class_name="pyreball-code-wrapper"):
        add_jquery = True
        groups_of_links_to_add.add("highlight_js")
    if _contains_class(html_text=html_content, class_name="pyreball-table-wrapper"):
        add_jquery = True
        groups_of_links_to_add.add("datatables")
    if _contains_class(html_text=html_content, class_name="pyreball-altair-fig"):
        groups_of_links_to_add.add("altair")
    if _contains_class(html_text=html_content, class_name="pyreball-plotly-fig"):
        groups_of_links_to_add.add("plotly")
    if _contains_class(html_text=html_content, class_name="pyreball-bokeh-fig"):
        groups_of_links_to_add.add("bokeh")

    # gather all links; jquery must be first
    links_to_add = "\n".join(
        (external_links["jquery"] if add_jquery else [])
        + [
            el
            for group in sorted(list(groups_of_links_to_add))
            for el in external_links[group]
        ]
    )
    html_content = re.sub("<!--PYREBALL_HEAD_LINKS-->", links_to_add, html_content)
    return html_content


def _insert_inline_highlight_script(html_content: str) -> str:
    if _contains_class(html_text=html_content, class_name="inline-highlight"):
        script_text = textwrap.dedent(
            """
        <script>
            $(document).ready(function () {
                $('code.inline-highlight').each(function (i, block) {
                    hljs.highlightBlock(block);
                });
            });
        </script>"""
        )
        html_content = re.sub(
            "<!--PYREBALL_INLINE_HIGHLIGHT_SCRIPT-->", script_text, html_content
        )
    else:
        html_content = re.sub(
            "<!--PYREBALL_INLINE_HIGHLIGHT_SCRIPT-->", "", html_content
        )

    return html_content


def _finish_html_file(
    html_path: Path, include_toc: bool, external_links: Dict[str, List[str]]
) -> None:
    """
    Load the printed HTML and finish substitutions to make it complete.

    Args:
        html_path: Path to the HTML file.
        include_toc: Whether to include the table of contents.
    """
    with open(html_path, "r") as f:
        lines = f.readlines()

    lines = _replace_ids(lines)
    lines = _insert_heading_title_and_toc(lines=lines, include_toc=include_toc)

    html_content = "".join(lines)
    html_content = _insert_js_and_css_links(html_content, external_links)
    html_content = _insert_inline_highlight_script(html_content)

    with open(html_path, "w") as f:
        f.write(html_content)


parameter_specifications = [
    IntegerParameter(
        "--page-width",
        boundaries=(40, 100),
        default=80,
        help=(
            "Width of the page container in percentage. "
            "An integer in the range 40..100. "
            "If set outside the range, the value will be shifted towards the range."
        ),
    ),
    ChoiceParameter(
        "--keep-stdout",
        choices=["yes", "no"],
        default="no",
        help="Print the output to stdout too.",
    ),
    ChoiceParameter(
        "--toc", choices=["yes", "no"], default="no", help="Include table of contents."
    ),
    ChoiceParameter(
        "--numbered-headings",
        choices=["yes", "no"],
        default="no",
        help="Number the headings.",
    ),
    ChoiceParameter(
        "--align-code-blocks",
        choices=["left", "center", "right"],
        default="center",
        help="Alignment of code blocks.",
    ),
    ChoiceParameter(
        "--code-block-captions-position",
        choices=["top", "bottom"],
        default="top",
        help="Position of the code block captions.",
    ),
    ChoiceParameter(
        "--numbered-code-blocks",
        choices=["yes", "no"],
        default="no",
        help="Number the code blocks.",
    ),
    ChoiceParameter(
        "--align-tables",
        choices=["left", "center", "right"],
        default="center",
        help="Alignment of tables.",
    ),
    ChoiceParameter(
        "--table-captions-position",
        choices=["top", "bottom"],
        default="top",
        help="Position of the table captions.",
    ),
    ChoiceParameter(
        "--numbered-tables",
        choices=["yes", "no"],
        default="no",
        help="Number the tables.",
    ),
    ChoiceParameter(
        "--tables-display-option",
        choices=["full", "paging", "scrolling"],
        default="full",
        help="How to display tables. Either full, with scrollbar, or with paging.",
    ),
    StringParameter(
        "--tables-paging-sizes",
        default="10,25,100,All",
        help=(
            "The paging sizes that can be selected. "
            "Allowed values are integers and string 'all' "
            "(no matter the case of letters), "
            "written as a non-empty comma-separated list. "
            "Ignored when tables-display-option is not 'paging'."
        ),
        validation_function=check_paging_sizes_string_parameter,
    ),
    StringParameter(
        "--tables-scroll-y-height",
        default="300px",
        help=(
            "Height of the tables when 'scrolling' display option is set. "
            "Any string compatible with CSS sizing can be used, "
            "e.g. '300px', '20em', etc. "
            "Ignored with other display options."
        ),
    ),
    ChoiceParameter(
        "--tables-scroll-x",
        choices=["yes", "no"],
        default="yes",
        help="Whether to allow horizontal scrolling on tables.",
    ),
    ChoiceParameter(
        "--sortable-tables",
        choices=["yes", "no"],
        default="no",
        help="Whether to make the tables sortable.",
    ),
    ChoiceParameter(
        "--tables-search-box",
        choices=["yes", "no"],
        default="no",
        help="Whether to show search box for tables.",
    ),
    StringParameter(
        "--tables-datatables-style",
        default="display",
        help=(
            "Datatables class(es) that affect the tables styling. "
            "If multiple classes are provided, "
            "separate them either with commas or spaces."
        ),
    ),
    ChoiceParameter(
        "--align-figures",
        choices=["left", "center", "right"],
        default="center",
        help="Alignment of figures.",
    ),
    ChoiceParameter(
        "--figure-captions-position",
        choices=["top", "bottom"],
        default="bottom",
        help="Position of the figure captions.",
    ),
    ChoiceParameter(
        "--numbered-figures",
        choices=["yes", "no"],
        default="no",
        help="Number the figures.",
    ),
    ChoiceParameter(
        "--matplotlib-format",
        choices=["png", "svg"],
        default="svg",
        help="Format of matplotlib figures.",
    ),
    ChoiceParameter(
        "--matplotlib-embedded",
        choices=["yes", "no"],
        default="no",
        help=(
            "Whether to embedded matplotlib figures directly into HTML. "
            "Only for svg format."
        ),
    ),
]


def _check_existence_of_config_files(
    config_dir_path: Path, recommendation_msg: str
) -> None:
    required_filename = [
        CONFIG_INI_FILENAME,
        LINKS_INI_FILENAME,
        STYLES_TEMPLATE_FILENAME,
        HTML_TEMPLATE_FILENAME,
    ]
    for filename in required_filename:
        if not (config_dir_path / filename).exists():
            raise FileNotFoundError(
                f"Config directory '{config_dir_path}' "
                f"does not contain all necessary files: "
                f"{', '.join(required_filename)}. {recommendation_msg}"
            )


def _get_config_directory(config_dir_path: Optional[Path] = None) -> Path:
    """Get the directory with the config files.

    If config path is provided by user, it is used primarily.
    If not, pyreball will try to check if config under home directory exists.
    If it does not exist, it will fall back to the default config directory
    in installation directory.
    In all cases, tbe function checks whether all config files exist.

    Args:
        config_dir_path: Optional path to the config file set by user
            through CLI argument. Can be a relative or absolute path
            and can contain ~ for home dir.

    Returns:
        An absolute path to the config directory.
    """
    rec_msg = "Try to regenerate the configs by pyreball-generate-config command."
    # Try to use the config directory from CLI argument
    if config_dir_path is not None:
        config_dir_path = config_dir_path.expanduser().resolve()
        if not config_dir_path.exists():
            raise NotADirectoryError(
                f"Provided config directory '{config_dir_path}' does not exist"
            )
        else:
            _check_existence_of_config_files(
                config_dir_path=config_dir_path, recommendation_msg=rec_msg
            )
            return config_dir_path

    # Try to use config directory from home directory if it exists
    home_config_dir_path = Path.home() / ".pyreball"
    if home_config_dir_path.exists():
        _check_existence_of_config_files(
            config_dir_path=home_config_dir_path, recommendation_msg=rec_msg
        )
        return home_config_dir_path

    # Fallback to the default config directory in installation directory
    default_config_dir_path = DEFAULT_PATH_TO_CONFIG
    _check_existence_of_config_files(
        config_dir_path=default_config_dir_path,
        recommendation_msg="Try re-installing pyreball.",
    )
    return default_config_dir_path


def _get_output_dir_and_file_stem(
    input_path: Path, output_path_str: Optional[Path]
) -> Tuple[Path, str]:
    """
    Obtain the output directory for the HTML file and output filename stem.

    Args:
        input_path: Path to the input script. Can be a relative or absolute path
            and can contain ~ for home dir.
        output_path_str: Optional path denoting the output path.
            Either as path to an HTML file or a directory.
            Can be a relative or absolute path and can contain ~ for home dir.

    Returns:
        A tuple of (output_dir_path, filename_stem) where output_dir_path
        is an absolute path to the output directory, and filename_stem
        is the output filename stem.
    """
    if not input_path.is_file():
        raise FileNotFoundError(f"File {input_path} does not exist.")

    if not output_path_str:
        # use the directory of the input file
        output_dir_path = input_path.parents[0].resolve()
        filename_stem = input_path.stem
    else:
        output_path = output_path_str.expanduser().resolve()
        if output_path.suffix == ".html":
            filename_stem = output_path.stem
            output_dir_path = output_path.parents[0]
        else:
            output_dir_path = output_path
            filename_stem = input_path.stem
        output_dir_path.mkdir(parents=True, exist_ok=True)

    return output_dir_path, filename_stem


class PathAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if values.strip() == "":
            raise argparse.ArgumentError(
                self, "Path argument cannot contain only whitespaces."
            )
        setattr(namespace, self.dest, Path(values))


def parse_arguments(args) -> Dict[str, Optional[Union[str, int]]]:
    parser = argparse.ArgumentParser(
        description=(
            "Generate Python report. "
            "All pyreball options must be specified before input-path argument. "
            "Any options or arguments after input-path are passed "
            "to the processed Python script. "
        )
    )
    for input_param in parameter_specifications:
        input_param.add_argument_to_parser(parser)
    parser.add_argument(
        "--output-path",
        help=(
            "Output path. Either path representing an HTML file or a directory. "
            "Any path with suffix different from '.html' will be considered "
            "a directory path. "
            "All parent directories are automatically created if they do not exist."
        ),
        action=PathAction,
    )
    parser.add_argument(
        "--config-path",
        help=(
            "Path to config directory. "
            "If not provided, Pyreball will try to use $HOME/.pyreball directory. "
            "If it does not exist, it will then try to use the default "
            "config directory in installation directory. "
        ),
        action=PathAction,
    )
    parser.add_argument(
        "input-path",
        help="Input file path. Must represent an existing Python script.",
        action=PathAction,
    )
    parser.add_argument(
        "script-args",
        nargs=argparse.REMAINDER,
        help="Remaining arguments that are passed to the Python script.",
    )
    variables = vars(parser.parse_args(args))
    # positional arguments must be renamed manually
    variables["input_path"] = variables["input-path"]
    variables["script_args"] = variables["script-args"]
    del variables["input-path"]
    del variables["script-args"]
    return variables


def main() -> None:
    args_dict = parse_arguments(sys.argv[1:])
    script_args_string = " ".join(cast(List[str], args_dict.pop("script_args")))
    input_path = cast(Path, args_dict.pop("input_path"))
    input_path = input_path.expanduser().resolve()
    output_path = cast(Optional[Path], args_dict.pop("output_path"))
    config_path = cast(Optional[Path], args_dict.pop("config_path"))

    output_dir_path, filename_stem = _get_output_dir_and_file_stem(
        input_path, output_path
    )
    # Directory, where HTML's images would be stored;
    # It basically contains both the output directory and HTML filename stem
    # in one value.
    html_dir_path_str = str(output_dir_path / filename_stem)
    html_path = output_dir_path / f"{filename_stem}.html"

    cli_parameters = check_and_fix_parameters(
        parameters=args_dict,
        parameter_specifications=parameter_specifications,
        none_allowed=True,
    )

    config_directory = _get_config_directory(config_path)
    file_config_parameters = get_file_config(
        filename=CONFIG_INI_FILENAME,
        directory=config_directory,
        parameter_specifications=parameter_specifications,
    )
    external_links = get_external_links_from_config(
        filename=LINKS_INI_FILENAME,
        directory=config_directory,
    )

    parameters = merge_parameter_dictionaries(
        primary_parameters=cli_parameters,
        secondary_parameters=file_config_parameters,
        parameter_specifications=parameter_specifications,
    )

    os.environ["_TMP_PYREBALL_GENERATOR_PARAMETERS"] = json.dumps(
        {**parameters, "html_dir_path": html_dir_path_str}
    )

    # remove the directory with images if it exists:
    carefully_remove_directory_if_exists(directory=Path(html_dir_path_str))

    css_definitions = get_css(
        filename=STYLES_TEMPLATE_FILENAME,
        directory=config_directory,
        page_width=cast(int, parameters["page_width"]),
    )

    html_begin, html_end = get_html(
        template_path=Path(
            pkg_resources.resource_filename("pyreball", f"cfg/{HTML_TEMPLATE_FILENAME}")
        ),
        title=filename_stem,
        css_definitions=css_definitions,
    )

    with open(html_path, "w") as f:
        f.write(html_begin)
    try:
        # Use {sys.executable} instead of just "python" command as it may not work
        # correctly as a PyCharm external tool
        os.system(f"{sys.executable} {input_path} {script_args_string}")
    finally:
        with open(html_path, "a") as f:
            f.write(html_end)

    _finish_html_file(
        html_path=html_path,
        include_toc=parameters["toc"] == "yes",
        external_links=external_links,
    )


if __name__ == "__main__":
    main()
